/**

Copyright ( 2001 The University of Kansas
COPYRIGHT AND LIMITATION ON USE
The documents and software (collectively the "Information") developed by the University of Kansas, Information and Telecommunication Technology Center (Hereinafter referred to as KU) is the property of KU or its licensors and is protected by copyright and other intellectual property laws. The information may be viewed, copied, run, modified or printed for educational and research purposes only and shall not be published, sold, broadcast, or redistributed, either directly or indirectly, in whole or in part, by any means or methods for commercial purposes. The Information may not be used for any purpose not expressly approved herein. Uses not approved include, but are not limited to, commercial distribution, in whole or in part.  Any rights not expressly granted herein are reserved by KU. 

Nothing contained herein shall be construed as conferring by implication, estoppel or otherwise any license or right under any patent or trademark of KU or any third party. Except as expressly provided above, nothing contained herein shall be construed as conferring any license or right under any KU copyright. 

DISCLAIMER OF WARRANTY
Use of the  information is at your sole risk. The information, is provided "AS IS" and without warranty of any kind and KU  AND ITS AFFILIATES, AGENTS AND LICENSORS , EXPRESSLY DISCLAIM ALL WARRANTIES, EXPRESS or IMPLIED, INCLUDING, BUT NOT LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND OF FITNESS FOR A PARTICULAR PURPOSE AND ANY EXPRESS OR IMPLIED WARRANTY AGAINST INFRINGEMENT. KU DOES NOT WARRANT THAT THE FUNCTIONS CONTAINED IN THE SOFTWARE WILL MEET YOUR REQUIREMENTS, OR THAT THE OPERATION OF THE SOFTWARE WILL BE UNINTERRUPTED OR ERROR-FREE, OR THAT DEFECTS IN THE SOFTWARE WILL BE CORRECTED. FURTHERMORE, KU DOES NOT WARRANT OR MAKE ANY REPRESENTATIONS REGARDING THE USE OR THE RESULTS OF THE USE OF THE SOFTWARE OR ITS DOCUMENTATION IN TERMS OF THEIR CORRECTNESS, ACCURACY, RELIABILITY, OR OTHERWISE. NO ORAL OR WRITTEN INFORMATION OR ADVICE GIVEN BY KU OR A KU AUTHORIZED REPRESENTATIVE SHALL CREATE A WARRANTY OR IN ANY WAY INCREASE THE SCOPE OF THIS WARRANTY. SHOULD THE SOFTWARE PROVE DEFECTIVE, YOU (AND NOT KU), ASSUMES THE ENTIRE COST OF ALL NECESSARY SERVICING, REPAIR OR CORRECTION. 

LIMITATION OF LIABILITY

NEITHER KU NOR ANY OF ITS AFFILIATES, AGENTS OR REPRESENTATIVES WILL BE LIABLE FOR ACTUAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES (INCLUDING LOSS OF PROFITS, LOSS OF BUSINESS OR SAVINGS), EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGES. IN NO EVENT WILL KU BE LIABLE FOR ANY ACTUAL, SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER, WHETHER IN AN ACTION OF CONTRACT, TORT OR ACTION OF ANY KIND, ARISING OUT OF OR IN CONNECTION WITH THE USE OF THE INFORMATION.  KU, ITS AFFILIATES, AGENTS AND REPRESENTATIVES WILL NOT BE LIABLE FOR NEGLIGENCE, RECKLESSNESS, INTENTIONAL MISCONDUCT OR ANY OTHER ACTS OR OMISSIONS OF THIRD PARTIES. 
GOVERNING LAW
This Agreement, all intellectual property issues, and your rights and obligations shall be governed by the laws of the United States of America and the State of Kansas, excluding their conflicts of laws rules. 


**/

package magician.Node;

import java.io.*;
import java.net.*;
import java.util.*;
import magician.Utils.*;
import magician.PerfTools.*;

  /**
   * Acknowledgement manager. It maintains the outgoing ports and the sequence
   * numbering on a per-port, per-type basis. This enables Smartpackets to put
   * in their own acknowledgement and sequencing schemes
   *
   * Starts an port thread for each outgoing port. Each thread has its own
   * waiting queue. 
   * @see PortThread
   * @see StopAndWaitAck
   * @see ReliableCommFW
   */
public class AckManager implements Runnable, PerfConstants {

  /**
   * table of ports on which acks are sent/received
   */
  static Hashtable Ports;
  static Hashtable NodesTable;
  static Hashtable PortThreadsTbl;
  static Hashtable SeqNumsTbl;
  static Queue outQ;

  /**
   * default timeout period for re-transmission
   */
  final static int TIMEOUT_PERIOD = 300;

  /**
   * pool of unused Ack packets
   */
  static UnsynchVector AckPool;

  /**
   * initial pool size
   */
  static int ACK_POOL_SIZE = 50;


  static EventManager EventMgr = null;

  /**
   * constructor
   * @param NodesTable table of neighboring nodes
   */
  public AckManager(Hashtable NodesTable) {
    this.NodesTable = NodesTable;
    SeqNumsTbl = new Hashtable();
    PortThreadsTbl = new Hashtable();
    outQ = new Queue();

    // set up a pool of ack packets
    int plsize = ACK_POOL_SIZE;
    AckPool = new UnsynchVector(plsize);

    while (plsize > 0) {
      Acknowledgement Ack = new Acknowledgement();
      AckPool.addElement(Ack);
      plsize--;
    }
   
    Ports = new Hashtable();
    EventMgr = (EventManager) ActiveNodeManager.HaveService("EventManager");
  }

  /**
   * returns the next available Ack packet
   * increases pool size if all exhausted
   */
  public synchronized static Acknowledgement getAckPacket() {
    Thread thr = Thread.currentThread();
    int prio = thr.getPriority();
    thr.setPriority(prio + 1);

    Acknowledgement Ack = null;

    try {
      Ack = (Acknowledgement) AckPool.firstElement();
      AckPool.removeElement(Ack);
      Ack.reinitialize();
    } catch (NoSuchElementException e) {
      // AckPool exhausted, increase capacity
      int cap = AckPool.capacity();

      // re-populate pool
      while (cap > 0) {
        Ack = new Acknowledgement();
        AckPool.addElement(Ack);
        cap--;
      }
 
      // save one for return
      Ack = new Acknowledgement();
      Ack.reinitialize();
    }

    thr.setPriority(prio);
    return Ack;
  }

  /**
   * returns the next sequence number for the type
   *
   * @param typeId the SmartPacket type
   * @return the next sequence number 
   */
  public static int getNextSequenceNumber(String PortName, String typeId) {
    
    String SeqCache = "__AMSeq." + PortName + "." + typeId; 

    int seqNo = 1;
 
    if (SeqNumsTbl.containsKey(SeqCache)) {
      seqNo = ((Integer) SeqNumsTbl.get(SeqCache)).intValue();
    }
      
    SeqNumsTbl.put(SeqCache, new Integer(seqNo+1));

    return seqNo;
  }

  /**
   * enqueue packets in appropriate port bucket
   * @param port name of neighboring active node 
   * @param SPObject the SmartPacket object
   * @param seqNo the sequence number for this packet
   * @param timeout the retransmission timeout
   */
  public synchronized static void queueForTransmission(String port, ReliableCommFW SPObject) throws NoRouteToHostException {
    Thread thr = Thread.currentThread();
    int prio = thr.getPriority();
    thr.setPriority(Thread.MAX_PRIORITY);
   
    if (Ports.containsKey(port) && Ports.get(port) == null) throw new NoRouteToHostException();

    outQ.insert(port);
    outQ.insert(SPObject);
    String type = new String(SPObject.getTypeId());
    outQ.insert(type);
    OnTheWirePacket OTWP = OnTheWirePacket.getOTWP(SPObject.PacketID);
    outQ.insert(OTWP);
    if (EventMgr.traceDepthSupported(TR_PACKET)) {
    String sbuf = "NODE=" + port + " AckQ.SIZE=" + String.valueOf(outQ.size());
    EventMgr.postEvent(TR_PACKET, "AckManager.Queue.Size", System.currentTimeMillis(), SPObject.PacketID, sbuf);
    }
    thr.setPriority(prio);
    return;
  }


  public void run() {

    String port = null;
    ReliableCommFW SPObject = null;
    OnTheWirePacket OTWP = null;
    String type = null;

    while (true) {
      try {
        port = (String) outQ.remove();
        SPObject = (ReliableCommFW) outQ.remove();
        type = (String) outQ.remove();
        OTWP = (OnTheWirePacket) outQ.remove();
      } catch (InterruptedException ie) {
        System.err.println("AckManager: interrupted");
        continue;
      }
   
      // System.err.println("AckManager: getting SPobject for port " + port);
      int seqNo = getNextSequenceNumber(port, type);
      SPObject.AckProto.seqNo = seqNo;

      SliWinPortThread pThread = null;
      String portkey = "__PortThread_" + port + type;

      /** check if port thread exists for this type **/
      if (PortThreadsTbl.containsKey(portkey)) {
        pThread = (SliWinPortThread) PortThreadsTbl.get(portkey);
      } else {
        // need to create port thread for this type
        Hashtable nodeProps = (Hashtable) NodesTable.get(port);
        String isNeighbor = (String) nodeProps.get("node.neighbor");
        if (isNeighbor.equals("yes")) {
          // System.err.println("Starting port thread for port " + port); 

          pThread = new SliWinPortThread(nodeProps);
          pThread.setPriority(Thread.MAX_PRIORITY);
          Ports.put(port, pThread);
          PortThreadsTbl.put(portkey, pThread);
          pThread.start();
        } else {
          System.err.println("AckManager: Cannot create port thread for non-neighbor: " + port);
        }
      }
   
      try {
        OTWP.setObject(SPObject);
        pThread.queueForTransmission(seqNo, OTWP.WritePacket());
        // System.err.println("AckManager: Setting Port " + port + " seq num = " + seqNo);      
        SPObject = null;
        OTWP = null;
      
      } catch (NullPointerException e) {
          if (pThread == null) 
            System.err.println("AckManager: No PortThread for " + port);
          else 
            e.printStackTrace();
      } catch (NoRouteToHostException nrhe) {
        Ports.put(port, null);
      } catch (IOException ioe) {
        ioe.printStackTrace();
      }
    }
  }


  /**
    * interface for Acknowledgement packet to send receipt to corresponding port
    * thread.
    * @param port name of active node associated with the port thread
    * @param seqNo the sequence number being acknowledged
    * @param typeId the type identifier of the Smartpacket
    * @exception IndexOutOfBoundsException if sequence number is out of order.
    */
  public static void confirmPacket(String port, int seqNo, String typeId) throws IndexOutOfBoundsException {
    SliWinPortThread pThread = null;
    String portkey = "__PortThread_" + port + typeId;

    try {
      if (PortThreadsTbl.containsKey(portkey)) {
        pThread = (SliWinPortThread) PortThreadsTbl.get(portkey);

        pThread.confirmPacket(seqNo);
      } else {
        System.err.println("AckManager: state " + portkey + " not found");
        System.exit(1);
      } 
    } catch (NullPointerException npe) {
      throw new IndexOutOfBoundsException();
    }
  } 
}
