/**

Copyright ( 2001 The University of Kansas
COPYRIGHT AND LIMITATION ON USE
The documents and software (collectively the "Information") developed by the University of Kansas, Information and Telecommunication Technology Center (Hereinafter referred to as KU) is the property of KU or its licensors and is protected by copyright and other intellectual property laws. The information may be viewed, copied, run, modified or printed for educational and research purposes only and shall not be published, sold, broadcast, or redistributed, either directly or indirectly, in whole or in part, by any means or methods for commercial purposes. The Information may not be used for any purpose not expressly approved herein. Uses not approved include, but are not limited to, commercial distribution, in whole or in part.  Any rights not expressly granted herein are reserved by KU. 

Nothing contained herein shall be construed as conferring by implication, estoppel or otherwise any license or right under any patent or trademark of KU or any third party. Except as expressly provided above, nothing contained herein shall be construed as conferring any license or right under any KU copyright. 

DISCLAIMER OF WARRANTY
Use of the  information is at your sole risk. The information, is provided "AS IS" and without warranty of any kind and KU  AND ITS AFFILIATES, AGENTS AND LICENSORS , EXPRESSLY DISCLAIM ALL WARRANTIES, EXPRESS or IMPLIED, INCLUDING, BUT NOT LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND OF FITNESS FOR A PARTICULAR PURPOSE AND ANY EXPRESS OR IMPLIED WARRANTY AGAINST INFRINGEMENT. KU DOES NOT WARRANT THAT THE FUNCTIONS CONTAINED IN THE SOFTWARE WILL MEET YOUR REQUIREMENTS, OR THAT THE OPERATION OF THE SOFTWARE WILL BE UNINTERRUPTED OR ERROR-FREE, OR THAT DEFECTS IN THE SOFTWARE WILL BE CORRECTED. FURTHERMORE, KU DOES NOT WARRANT OR MAKE ANY REPRESENTATIONS REGARDING THE USE OR THE RESULTS OF THE USE OF THE SOFTWARE OR ITS DOCUMENTATION IN TERMS OF THEIR CORRECTNESS, ACCURACY, RELIABILITY, OR OTHERWISE. NO ORAL OR WRITTEN INFORMATION OR ADVICE GIVEN BY KU OR A KU AUTHORIZED REPRESENTATIVE SHALL CREATE A WARRANTY OR IN ANY WAY INCREASE THE SCOPE OF THIS WARRANTY. SHOULD THE SOFTWARE PROVE DEFECTIVE, YOU (AND NOT KU), ASSUMES THE ENTIRE COST OF ALL NECESSARY SERVICING, REPAIR OR CORRECTION. 

LIMITATION OF LIABILITY

NEITHER KU NOR ANY OF ITS AFFILIATES, AGENTS OR REPRESENTATIVES WILL BE LIABLE FOR ACTUAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES (INCLUDING LOSS OF PROFITS, LOSS OF BUSINESS OR SAVINGS), EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGES. IN NO EVENT WILL KU BE LIABLE FOR ANY ACTUAL, SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER, WHETHER IN AN ACTION OF CONTRACT, TORT OR ACTION OF ANY KIND, ARISING OUT OF OR IN CONNECTION WITH THE USE OF THE INFORMATION.  KU, ITS AFFILIATES, AGENTS AND REPRESENTATIVES WILL NOT BE LIABLE FOR NEGLIGENCE, RECKLESSNESS, INTENTIONAL MISCONDUCT OR ANY OTHER ACTS OR OMISSIONS OF THIRD PARTIES. 
GOVERNING LAW
This Agreement, all intellectual property issues, and your rights and obligations shall be governed by the laws of the United States of America and the State of Kansas, excluding their conflicts of laws rules. 


**/
package magician.Node;

import java.lang.*;
import java.io.*;
import java.util.*;
import magician.PerfTools.*;
import magician.Utils.*;

/**
 * The incoming SmartPacket consists of an outer frame and an inner frame. The outer frame consists
 * of the Service Selector field which indicates the processing environment that is to be used to execute 
 * the SmartPacket. The primary function of the Demultiplexer class is to read the Service Selector field 
 * and use it to route the SmartPacket to the correct processing environment. The Demultiplexer runs a
 * a thread and reads the packets from a queue which is filled in by the PortManager.
 * The processing environment IDs (corresponding to the values of the Service Selector fields)  and paths 
 * to the executables are stored on the Four11 information server. At boot time, the Active Node manager 
 * reads this information and initializes the Demux with all the processing environments. 
 */


public class Demultiplexer implements Runnable, ActiveNetsConstants, PerfConstants {
  /**
   * queue for incoming packets
   */
  public static Queue IncomingSPs_Q; // the queue on which the PortManager queues incoming packets
  /**
   * processing environemnts list
   */
  private Hashtable ProcEnvsTable;  // table to find the processing environments indexed by TypeID
  
  /**
   * the bytes comprising the ANEP-encapsulated SmartPacket
   */
  private byte[] packetBytes;

  /**
   * stream to use for reading bytes 
   */
  private DataInputStream bufstream;

  /**
  * Event Manager forlogging events
  */
  private static EventManager EventMgr;

  /**
   * packet count per logging cycle. This is a cyclic counter which counts upto
   * PACKETCOUNT packets, logs the time it takes to reach the value and
   * is reset
   */
  public static int PACKETCOUNT = 10;
  public static int numPackets = 0;
  public static int numDiscards = 0;     /** is not cyclic */
  public static int numForwarded = 0;    /** is not syclic */

  /**
   * this node's name
   */
  private String nodeName;

  /**
   * initializes the Demux 
   * @param ProcEnvs  all processing environments available at this node
   */
  public Demultiplexer(Hashtable ProcEnvs) {
    IncomingSPs_Q = new Queue(20);
    ProcEnvsTable = ProcEnvs; 
    nodeName = ActiveNode.GetNodeName();
    /**
     *  initialize an event manager to log events
     */
    EventMgr = (EventManager) ActiveNodeManager.HaveService("EventManager");
  }

  /**
   * gets a handle to the log device for reading 
   * @return an event manager object
   */
  protected EventManager getEventManager() {
    return EventMgr;
  }

  /** 
   * define a new type processing environment to the Demux as a <type, pathname> pair.
   * When processing a SmartPacket, its type ID is used to get the path to the processing 
   * environment and the new environment is exec'ed.
   * @param typeID  the type of packet which needs processing
   * @param procEnv  an Object representing the processing environment
   */
  protected void defineTypeProcessor(byte[] typeID, Object procEnv) {
    // enter the processing function into the table indexed by typeID
    ProcEnvsTable.put(typeID, procEnv);
  }

  /**
   * performs initialization steps so that the individual fields of the incoming packet 
   * can be read
   */
  protected void initStream() {
    bufstream = new DataInputStream(new UnsynchByteArrayInputStream(packetBytes));
  }


  /**
    * called by PortManager to deposit newly arrived packet stream
    * 
    * @param port the port on which the packet arrived
    * @param packet the packet stream
    */
  protected static synchronized void queuePacket(int port, byte[] packet) {
    IncomingSPs_Q.insert((Integer) new Integer(port));
    IncomingSPs_Q.insert((Long) new Long(System.currentTimeMillis()));
    IncomingSPs_Q.insert((byte[]) packet);
    if (EventMgr.traceDepthSupported(TR_SYSTEM)) {
      String sbuf = "INPUT_PORT=" + String.valueOf(port);
      EventMgr.postEvent(TR_SYSTEM, "PortManager.PacketReceive.QueueinDemux", System.currentTimeMillis(), packet.length, sbuf);
    }
  }

  /**
   *
   * check for the version of the ANEP packet
   * @return true if version is correct, false otherwise
   */
  protected boolean checkVersion() 
  throws StreamCorruptedException {
    try {
      bufstream.reset();
      byte ver = bufstream.readByte(); 
      return true;
    } catch (EOFException e) {
      throw new StreamCorruptedException();
    } catch (IOException e) {
      throw new StreamCorruptedException();
    }
  }

  /**
   * returns the type ID of the packet
   * @return  the type id as an int
   */
  protected int getTypeID() 
  throws StreamCorruptedException {
    int typeid = 0;
    try {
      // skip to get to the beginning of the typeid field
      bufstream.reset();
      bufstream.skip(2);  // skip 2 bytes

      // type id field is 16 bits long
      typeid = (int) bufstream.readShort();
    } catch (EOFException e) {
      throw new StreamCorruptedException();
    } catch (IOException e) {
      throw new StreamCorruptedException();
    }
    return typeid;
  }

  /**
   * checks the flags field at the beginning of the byte stream to see if packet has to be 
   * discarded or forwarded if type is unknown at the current node
   * @return true if forwarding is enabled, false otherwise
   */
  protected boolean getForwardUnknownPacketFlag() 
  throws StreamCorruptedException {
    try {
      // skip the version field to get to the flags field
      bufstream.reset();
      bufstream.skip(1);

      byte flags = bufstream.readByte();

      return ((flags >> 6) & 0x1) == 0 ? true : false;
    } catch (EOFException e) {
      throw new StreamCorruptedException();
    } catch (IOException e) {
      throw new StreamCorruptedException();
    }
  }


  /**
   * goes through the TLV options to obtain the embedded address
   * @param addrType type of address, whether source or destination 
   * @return a byte array containing the forwarding address
   */
  protected byte[] getAddress(int addrType) 
  throws StreamCorruptedException {
  
    byte[] address = null;
    try {
      // skip the first 4 bytes to get to the header length field
      bufstream.reset();
      bufstream.skip(4);

      // header length field is 16 bits long and specifies the length in 32 bit words
      int headerLength = (int) bufstream.readShort();
      headerLength *= 4; // we save hdr len in bytes

      // skip the packetlength field  (2 bytes)
      bufstream.skip(2);
  
      short optionType = 0;  // undefined option type
      while(optionType != addrType) {

        // read the next TLV option header
        byte[] optField = new byte[2];
        bufstream.readFully(optField, 0, 2);
        // extract the flag and option type bits
        byte flags = (byte) ((optField[0] & 0xC0) >> 6); // get the top two flag bits
        short hiBits = (short) (optField[0] & 0x3F);
        short loBits = (short) (optField[1] & 0xFF);
     
        optionType = (short) ((hiBits << 8) + (loBits << 0));

        int optionLength = (int) bufstream.readShort();
        optionLength *= 4; // since length of option is in 32 bit words

        if (optionType == addrType) {
          // read the scheme identifier to determine type of address
          int schemeID = (int) bufstream.readInt();

          int addressLen = 0;
          switch (schemeID) {
              // scheme ID determines what type of address we are reading
            case 1:
              addressLen = 4;
              break;
            case 2:
              addressLen = 16;
              break;
            case 3:
              addressLen = 6;
              break;
          }

          // read the payload to extract the destination address
          address = new byte[addressLen];
          bufstream.readFully(address, 0, address.length); 
          break;
        }
        else {
          // skip over this option
          bufstream.skip(optionLength - 4);
        }
      }
    } catch (EOFException e) {
      throw new StreamCorruptedException();
    } catch (IOException e) {
      throw new StreamCorruptedException();
    }
    
    return address;
  }

  /**
   * converts a byte array representing an address into an IP
   * address string of the form "%d.%d.%d.%d"
   * @param a byte array containing the IP address
   * @return a String of the form "%d.%d.%d.%d"
   */
  private String convertToString(byte[] IPAddr) {
    String IPString = null;
      int addr0 = (int) (IPAddr[0] & 0xFF);
      int addr1 = (int) (IPAddr[1] & 0xFF);
      int addr2 = (int) (IPAddr[2] & 0xFF);
      int addr3 = (int) (IPAddr[3] & 0xFF);

      return (addr0 + "." + addr1 + "." + addr2 + "." + addr3);
  }


  /**
   * private method to log events to Event manager
   *
   * @param evType  a String identifying the event being logged
   * @param evValue value to be logged
   */
  void logEvent(String evType, long evValue) {
    if (EventMgr.traceDepthSupported(TR_SYSTEM)) {
      String sbuf = "VALUE=" + String.valueOf(evValue);
      EventMgr.postEvent(TR_SYSTEM, evType, System.currentTimeMillis(), 0, sbuf);
    }
  }

  /**
   * The run() method reads the packets off the queue, uses the TypeID field to index into the
   * ProcessingEnvs table to find the environment to call to run the SmartPacket in
   */
  public void run() {
    // loop reading packets
    int port = 0;
    String forwAddress = null;
    while (true) {
      // read from queue, blocks if queue is empty      
      try {
        port = (int) ((Integer) IncomingSPs_Q.remove()).intValue();
        long timeQueued = (long) ((Long) IncomingSPs_Q.remove()).longValue();
        packetBytes = (byte[]) IncomingSPs_Q.remove();
        /**
	 * increment packet count
	 */
        numPackets++;

	// SB commented out
        //if (numPackets > PACKETCOUNT) {
          // /** log time to event manager */
          // logEvent("Demultiplexer.ThroughputStats.PacketsReceived", (long) numPackets);
					//EventMgr.wakeup();
          // PACKETCOUNT += 10;
        // }
        /**
	 * log time spent in queue
	 */

        logEvent("Demultiplexer.DelayStats.QueueingDelay", System.currentTimeMillis() - timeQueued);

        // set up the byte stream so that we can read individual fields
        initStream();
 
        // parse TLV options to find out destination address for forwarding
        // for the time being, we will use source address and send the 
        // packet back to its sender, if its type is unrecognized at this node.
        forwAddress = convertToString(getAddress(TLV_SRCIDENT));

        // log type and source address to event log
        if (EventMgr.traceDepthSupported(TR_SYSTEM)) {
          String sbuf = "FORW_ADDR=" + forwAddress + " PACKETSIZE=" + String.valueOf(packetBytes.length);
          EventMgr.postEvent(TR_SYSTEM, "Demultiplexer.Info.PacketSource", System.currentTimeMillis(), 0, sbuf);
        }
        // check TypeID and see if processing env exists for that type
        int envType = getTypeID();

        // dispatch packet to the processing function for the packet using the type id of the packet
        ExecEnvironment ExecEnv = (ExecEnvironment) ProcEnvsTable.get(new Integer(envType));
        if (ExecEnv == null) throw new UnknownTypeIDException();

        ExecEnv.Process(packetBytes);
        packetBytes = null;
      } catch (ClassCastException e) {
        logEvent("Demultiplexer.ThroughputStats.PacketLost", (long) 0);
        continue;
      } catch (StreamCorruptedException e) {
        // discard the packet
        numDiscards++;
        logEvent("Demultiplexer.ThroughputStats.PacketsDiscarded", (long) numDiscards);
        // fetch next packet from queue
        continue;
      } catch (InterruptedException e) {
        // fetch next packet from queue
        logEvent("Demultiplexer.ThroughputStats.PacketsLost", (long) numDiscards);
        continue;
      } catch (NoSuchElementException e) {
          continue;
      } catch (UnknownTypeIDException e) {
        try {
	  // type is unknown, check flag to find out what is to be done
          if (getForwardUnknownPacketFlag() == false) {
            // so packet discarding enabled
            throw new StreamCorruptedException();
          }
          numForwarded++;

          // send packet on its way
          ActiveNode.send(forwAddress, port, packetBytes);
        
          logEvent("Demultiplexer.ThroughputStats.PacketsForwarded", (long) numForwarded);
        } catch (StreamCorruptedException sc) {
          numDiscards++;
          logEvent("Demultiplexer.ThroughputStats.PacketsDiscarded", (long) numDiscards);
          // fetch next packet from queue
          continue;
	}
      }
    }
  }
}
