/*******************************************************************************
 * Copyright (c) 1997 by the University of Kansas Center for Research, Inc.
 * All Rights Reserved.  This software was developed by the Information Technology
 * and Telecommunication Center (ITTC) at the University of Kansas
 * with the assistance of DARPA.
 *
 * Permission to use, copy, modify and distribute this software and its
 * documentation for personal or educational use is hereby granted, provided
 * that copies of the software are not made or distributed for profit or
 * commercial advantage; that both the copyright notice and this permission
 * notice appear in all copies of the software, derivative works
 * or modified versions, and any portions thereof; that both notices appear in
 * supporting documentation; and that the use of this software is acknowledged
 * in any publications that might result from the use of this software.
 *
 * ITTC ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION, AND
 * DISCLAIMS ANY LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING
 * FROM THE USE OF THIS SOFTWARE.
 *
 * RCS Identification : $Id: KU_SmartPacket_V2.java,v 1.2 2001/05/20 15:30:11 kulkarni Exp kulkarni $
 *
 * @author Amit B. Kulkarni
 */


package magician.Node;
import java.lang.*;
import java.io.*;
import java.net.*;
import java.util.*;
import java.lang.reflect.*;
import magician.Node.PortInterface;
import magician.Node.Scheduler;
import magician.Node.RIPRouting;

/**
 *
 * This is the base class for all communication in the active network.  Any 
 * new SmartPacket that the user creates must extend this class.
 * This class represents the interface to the node resources. All access to 
 * node resources are governed through this class. In this release, access to
 * small state at an active node is not governed through this class. Use
 * <tt>ActiveNodeManager.CreateSmallState()</tt>, <tt>ActiveNodeManager.GetSmallState()</tt>,
 * and <tt>ActiveNodeManager.SetSmallState()</tt> primitives for such access.
 * However, it must be noted that those primitives will be rolled in with this
 * interface in the next release.
 * SmartPackets that are derived directly from this class are not guaranteed
 * reliable communication. To use reliable communication, use ReliableCommFW class.
 *
 * This class provides the entry point for all user code. The user code must be 
 * inserted by overriding the <tt>exec()</tt> method of this class. If one wishes
 * to build a complex protocol hierarchy before the user code is executed, the
 * <tt>build()</tt> method must be overridden with the appropriate framework code.
 * See <tt>ReliableCommFW</tt> for an example of a framework implementing reliable
 * communication.
 *
 * @see ReliableCommFW
 */
public abstract class KU_SmartPacket_V2  implements Runnable, Serializable {

  /**
   * the cpu cycles required by htis packet (currently ignored)
   */
  protected long CPU_Requirements = 0;

  /**
   * the memory in bytes consumed by this packet during execution (currently not recorded)
   */
  protected int Memory_Requirements = 0;

  /**
   * the network bandwidth requirements in bits/second (not used)
   */
  protected int IO_BW_Requirements = 0;

  // Address Information
  /**
   * name of source active node
   */
  protected String Source_Address;
  /**
   * name of destination active node
   */
  public String Destination_Address;

  /**
   * Packet ID, used by ActiveNodeManager to track packet
   */
  transient protected int PacketID = 0;

  /**
   * name of next hop active node
   */
  transient protected  String NextHop;

  /**
   * number of retries to get next hop from routing method
   */
  final static int MAX_RETRIES = 2;


  /**
   * the type identifier for this Smartpacket.
   */
  protected byte[] TypeID;

  transient protected String thisNode = null;


  transient protected boolean halted = false;

  /**
   * method called by Node to execute the Smartpacket
   */
  public final void run () { 
    thisNode = ActiveNodeManager.GetNodeName();
    // get exec env for packet thread
    SPThread myTh = SPThread.currentSPThread();
    SPExecEnv ee = myTh.parent;

    // notify exec env that thread is starting
    ee.notifyExecEnv();
 
    // build the framework for the protocol hierarchy
    build();
    stop();
  }

  /**
    * returns the type identifier for the SmartPacket
    * @return byte array representing the type identifier
    */
  public byte[] getTypeId() {
    return TypeID;
  }
 
  /**
    * gives the name of the active node from which this Smartpacket
    * was sent
    * @return a string representing the name of the source node
    */
  public String getSourceAddress() {
    return Source_Address;
  }
  
  /**
    * returns the canonical name of the active node if it exists,
    * else returns the gateway node from which the node can be
    * reached  
    * @param nodeName the name of the Active Node
    * @return name of active node, if it exists, else name of gateway node 
    */
  public String GetCanonicalAddress(String nodeName) {
    return PortInterface.getANAddress(nodeName);
  }
 
  /**
    * gives the base address of the active node, default is IP address
    * @param name of the active node
    * @return a string representing base address of active node
    */
  public String GetLinkLayerAddress(String nodeName) {
    return (String) PortInterface.getLLAddress(nodeName);
  }

  /**
    * gives the name of this active node 
    * @return string respresenting name of this node
    */
  public String GetNodeName() {
      return thisNode;
  }

  /**
    * gives the name of the network to which this node belongs
    * @return String representing network name
    */
  public String GetNetName() {
    return ActiveNodeManager.GetNetName();
  }

  /**
    * returns a service registered at the active node
    *
    * @param service name of the service
    * @return object representing service if it exists, else null
    */
  public Object GetService(String service) {
    return ActiveNodeManager.HaveService(service);
  }

  /**
    * returns the name of the active node that is the next hop towards the 
    * destination
    * @return name of active node 
    */
  public String GetNextHop() {
    int retries = 0;

    do {
      try {
        NextHop = RoutingMgr.defaultRouting().getnexthop(Destination_Address);
      } catch (NoRouteToHostException e ) {
        System.err.println("destination = " + Destination_Address + " Next Hop = " + NextHop);
        retries++;
        if (retries < MAX_RETRIES) {
	try {
         int intvl = ((RIPRouting) RoutingMgr.defaultRouting()).getUpdateInterval();
         SPThread.currentSPThread().sleep(intvl);
	} catch (InterruptedException ie) {
	}
        } else {
           pathNotFound();
        }
      }
    } while (NextHop == null);

    return NextHop;
  }

  /**
   * gives the number of hops required to reach a 
   * particular node from this node
   * 
   * @return number of hops required
   *
   * @exception NoRouteToHostException if no route exists
   */
  public int GetNumberOfHops(String node) throws NoRouteToHostException {
    return ((RIPRouting) RoutingMgr.defaultRouting()).getNumberOfHopsToNode(node);
  }
 
  /**
    * sends this Smartpacket towards destination
    */
  public void SendSmartPacket() {
      if (! thisNode.equals(Destination_Address)) {
        GetNextHop();
        // System.err.println("KU_SmartPacket_V2: next hop is " + NextHop);
        ActiveNodeManager.SendSmartPacket(NextHop, this);    
      }
  }

  /**
   * To be used when packet is to be sent directly without
   * execution at the local node
   * @param destination the name of the destination active node
   */
  public void SendSmartPacket(String destination) {
    Destination_Address = destination;
    SendSmartPacket();
  }

  /**
   * To be used when smartpacket is expected
   * to execute at this node
   * 
   * @param destination the ultimate destination of the packet
   */
  public void SendForProcessing(String destination) {
    Destination_Address = destination;
    ActiveNodeManager.ProcessSP(destination, this);
  }
  
  /**
    * interface that is to be used to build frameworks beyond the basic
    * unreliable delivery provided by this class
    * @see ReliableCommFW
    */
  public void build () {
    // communication framework designers must extend this method
    exec();

    SendSmartPacket();
  }

  /**
    * this method is overridden by user to implement user logic
    */
  public void exec() {
    // leaf classes must extend this method
  }

  /**
   * describes the components of the type class to the serilaized stream
   * This is called by user frameworks that are derived from this class
   * to describe related components that form the same type
   * @param components an array of Field objects that describe related classes
   * @param out ObjectOutputStream that is provided by WriteObject method during
   * serialization
   * @exception IOException if object cannot be serialized
   */
  protected void describeComponents(Field[] components, ObjectOutputStream out) throws IOException {
    
    // write the definitions for the transient classes which are part of
    // this framework.
    int count = components.length;
    for (int ix = 0; ix < count; ix++) {
      // obtain the class of the field
      Class type = components[ix].getType();
      // System.err.println("KU_SmartPacket_V2: Annotating component " + components[ix].getName() + " of type " + type.getName());
      ((SPObjectOutputStream) out).annotateClass(type);
    }
    
  }

  /**
    * stops execution of SmartPacket
    */
  protected void stop() {
    if (halted) {
	return;
    }
    halted = true;

    if (PacketID != 0) {
      OnTheWirePacket.destroyOTWP(PacketID);
    }
    Scheduler.invoke();
  }

  protected void halt() {
      Destination_Address = thisNode;
  }


  /**
    * sleep for specified amount of time
    * @param time amount of time to sleep
    */
  protected void sleep(long time) {
    try {
      SPThread.currentSPThread().sleep(time);
    } catch (InterruptedException ie) {
      ie.printStackTrace();
      halt();
    }
  }

  /**
    * relinquish the CPU 
    */
  protected void yield() {
    SPThread.currentSPThread().yield();
  }

  /**
    * enables small state to be created by user packet
    * @param SSName name identifier of small state
    */
  protected void CreateSmallState(String SSName) throws StateExistsException {
    ActiveNodeManager.CreateSmallState(SSName);
  }

  /**
    * synchronously acquires small state
    *
    * @param SSName  name identifier of small state
    * @return Object stored in small state
    * @exception StateNotFoundException if named small state does not exist
    * @exception StateEmptyException if state exists but is empty
    */
  protected Object GetSmallState(String SSName) throws StateNotFoundException, StateEmptyException {
    return ActiveNodeManager.GetSmallState(SSName);
  }

  /**
    * synchronously acquires small state but user can specify maximum time period
    *
    * @param SSName  name identifier of small state
    * @return Object stored in small state
    * @exception StateNotFoundException if named small state does not exist
    * @exception StateEmptyException if state exists but is empty
    */
  protected Object GetSmallState(String SSName, long timeout) throws StateNotFoundException, StateEmptyException {
    return ActiveNodeManager.GetSmallState(SSName, timeout);
  }

  /**
    * synchronously acquires small state
    *
    * @param SSName  name identifier of small state
    * @return Object stored in small state
    * @exception StateNotFoundException thrown if named small state does not exist
    */
  protected Object AcquireSmallState(String SSName) throws StateNotFoundException {
    Object state = null;

    while (state == null) {
      try {
        state = ActiveNodeManager.GetSmallState(SSName);
      } catch (StateEmptyException nsee) {
	SPThread.currentSPThread().yield();
      } 
    }

    return state;
  }

  /**
    * sets the contents of the named small state
    * @param SSName name identifier of small state
    * @param value contents to be placed
    * @exception StateNotFoundException if named small state is not found
    */
  protected void SetSmallState(String SSName, Object value) throws StateNotFoundException {
    ActiveNodeManager.SetSmallState(SSName, value);
  }

  /**
    * destroys previously created small state
    * @param SSName name identifier of small state
    */
  protected void DestroySmallState(String SSName) throws StateNotFoundException {
    ActiveNodeManager.DestroySmallState(SSName);
  }

  /**
    * called when node cannot find route to destination
    */
  protected void pathNotFound() {
    System.err.println("Cannot find route to " + Destination_Address + ", halting..");
    halt();
  }
}
