/**

Copyright ( 2001 The University of Kansas
COPYRIGHT AND LIMITATION ON USE
The documents and software (collectively the "Information") developed by the University of Kansas, Information and Telecommunication Technology Center (Hereinafter referred to as KU) is the property of KU or its licensors and is protected by copyright and other intellectual property laws. The information may be viewed, copied, run, modified or printed for educational and research purposes only and shall not be published, sold, broadcast, or redistributed, either directly or indirectly, in whole or in part, by any means or methods for commercial purposes. The Information may not be used for any purpose not expressly approved herein. Uses not approved include, but are not limited to, commercial distribution, in whole or in part.  Any rights not expressly granted herein are reserved by KU. 

Nothing contained herein shall be construed as conferring by implication, estoppel or otherwise any license or right under any patent or trademark of KU or any third party. Except as expressly provided above, nothing contained herein shall be construed as conferring any license or right under any KU copyright. 

DISCLAIMER OF WARRANTY
Use of the  information is at your sole risk. The information, is provided "AS IS" and without warranty of any kind and KU  AND ITS AFFILIATES, AGENTS AND LICENSORS , EXPRESSLY DISCLAIM ALL WARRANTIES, EXPRESS or IMPLIED, INCLUDING, BUT NOT LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND OF FITNESS FOR A PARTICULAR PURPOSE AND ANY EXPRESS OR IMPLIED WARRANTY AGAINST INFRINGEMENT. KU DOES NOT WARRANT THAT THE FUNCTIONS CONTAINED IN THE SOFTWARE WILL MEET YOUR REQUIREMENTS, OR THAT THE OPERATION OF THE SOFTWARE WILL BE UNINTERRUPTED OR ERROR-FREE, OR THAT DEFECTS IN THE SOFTWARE WILL BE CORRECTED. FURTHERMORE, KU DOES NOT WARRANT OR MAKE ANY REPRESENTATIONS REGARDING THE USE OR THE RESULTS OF THE USE OF THE SOFTWARE OR ITS DOCUMENTATION IN TERMS OF THEIR CORRECTNESS, ACCURACY, RELIABILITY, OR OTHERWISE. NO ORAL OR WRITTEN INFORMATION OR ADVICE GIVEN BY KU OR A KU AUTHORIZED REPRESENTATIVE SHALL CREATE A WARRANTY OR IN ANY WAY INCREASE THE SCOPE OF THIS WARRANTY. SHOULD THE SOFTWARE PROVE DEFECTIVE, YOU (AND NOT KU), ASSUMES THE ENTIRE COST OF ALL NECESSARY SERVICING, REPAIR OR CORRECTION. 

LIMITATION OF LIABILITY

NEITHER KU NOR ANY OF ITS AFFILIATES, AGENTS OR REPRESENTATIVES WILL BE LIABLE FOR ACTUAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES (INCLUDING LOSS OF PROFITS, LOSS OF BUSINESS OR SAVINGS), EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGES. IN NO EVENT WILL KU BE LIABLE FOR ANY ACTUAL, SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER, WHETHER IN AN ACTION OF CONTRACT, TORT OR ACTION OF ANY KIND, ARISING OUT OF OR IN CONNECTION WITH THE USE OF THE INFORMATION.  KU, ITS AFFILIATES, AGENTS AND REPRESENTATIVES WILL NOT BE LIABLE FOR NEGLIGENCE, RECKLESSNESS, INTENTIONAL MISCONDUCT OR ANY OTHER ACTS OR OMISSIONS OF THIRD PARTIES. 
GOVERNING LAW
This Agreement, all intellectual property issues, and your rights and obligations shall be governed by the laws of the United States of America and the State of Kansas, excluding their conflicts of laws rules. 


**/
package magician.Node;

import java.lang.*;
import java.io.*;
import java.util.*;
import java.util.zip.*;
import java.net.*;
import magician.Utils.*;
import magician.PerfTools.*;

/**
 * Holds the contents of a single packet received on the wire.
 * Most of the information is held inside the ANEPPacket which is
 * encapsulated inside it. But it also contains data and methods for
 * storage of local environment information and information regarding
 * the format of the KU SmartPacket
 *
  * @see SmartPacketLoader
 */

 class OnTheWirePacket implements ActiveNetsConstants, PerfConstants {

  /**
   * Class-wide packet id counter, gives the number of packets created
   */
  static int numPackets = 0; 

  /**
   * cache to hold the OnTheWirePackets
   */
  public static Hashtable OTWPcache = new Hashtable();

  /**
   * reference to the bytes of the SmartPacket  hot off-the-wire
   */
   // byte[] packetStream;

  /**
   * the version of this format (default 1)
   */
  short version = 1;       

  /**
   * flags 
   */
  byte flags;        

  /**
   * the environment this packet is processed in
   */
  byte[] EnvID;  

  /**
   * the type id of this packet. It is stored in one of the TLVs
   */
  byte[] TypeID = null;

  /**
   * the length of the ANEP header
   */
  int ANEPheaderLength;     

  /**
   *  the length of the packet
   */
  int packetLength;    

  /**
   * a vector containing the TLV options 
   */
  UnsynchVector TLVOptions;    

  /**
   * the payload of the ANEP packet
   */
  // byte[] payload;      
 
  /**
   * a table of names of the main and embedded classes carried by the ANEPPacket
   */
  protected Hashtable ClassNames;   
 
  /**
   * reference to the serialized object
   */
  protected KUSmartPacketV2 SPObject;

  /**
    * Reference to TLV containing object
    */
  private TLVOption ObjectTLV = null;

  /**
   * id given to the current packet
   */
  protected int PacketID = 0;

  /**
   * hash of the object data contained in the ANEPPacket computed when packet is received.
   * During transmission, this value is checked to see if it is the same as the initial value
   * if it is, the saved wirestream is sent instead of going through serialization
   * IMPORTANT: you cannot use the hashCode() method of Class Object to compute hash. This
   * is because the hash for the object remains the same even if the internal data changes
   * So we take the bytes of the object and use simple recursive hashing to get a
   * 16-bit hash code.
   * @see #ObjectHash
   */
  int InitialHash = 0; 
  int OutgoingHash;

  /** 
    * Eventmanager to log events
    */
  static EventManager EventMgr;

  /**
    * cache in which SmartPacket loaders for each TypeID are kept
    * @see SmartPacketLoader
    */
  public static Hashtable LoaderCache = new Hashtable();

  SmartPacketLoader SPLoader = null;
  String NodeName = null;

  /**
   * takes a ANEPPacket received over the network and creates a new OnTheWirePacket
   *
   * @param packetbytes the bytes of the incoming packet
   * @exception StreamCorruptedException if the bytes of the packet do not conform to the standard
    */
   protected OnTheWirePacket (byte[] packetbytes) 
    throws StreamCorruptedException {
    init();
    // packetStream = packetbytes;
    // find out time it takes to assemble a packet
    long assemblyStart = System.currentTimeMillis();

    assemblePacket(packetbytes);
    packetbytes = null;
    
    if (EventMgr.traceDepthSupported(TR_MEASURE)) {
      String sbuf = "VALUE=" + String.valueOf(System.currentTimeMillis() - assemblyStart);
      EventMgr.postEvent (TR_MEASURE, "ActiveNodeManager.PacketAssembly.Interval",                        System.currentTimeMillis(), (long)PacketID, sbuf);
    }
			    
    // set the id for the executable object to correlate the object to this OTW packet
    SetObjectID(PacketID);
  }

  /**
   * creates a new packet for transmission
   * @param envID the environment in which this packet is to be executed
   * @param typeId the type id of the packet
   * @param SPObject the SmartPacket object
   * @exception IllegalArgumentException if either of the arguments is not valid
   */
  protected OnTheWirePacket (int envID, byte[] typeId, KUSmartPacketV2 SPObject) 
      throws IllegalArgumentException, IOException {
    
    init();
    flags = (byte) 0x80;  // set forwarding flag
    // convert env id to bytes
    byte[] env = new byte[2];
    env[0] = (byte) ((envID >> 8) & 0xFF);
    env[1] = (byte) ((envID >> 0) & 0xFF);
    EnvID = env;
    TypeID = typeId;
    
    ANEPheaderLength = 8;  // minimum because no TLVs have in written in yet
    packetLength = 8;
    addTypeDef();
    SPLoader = getLoader(TypeID);
  }

  /**
   * common initialization routine for all constructors
   */
  private synchronized void init() {
    numPackets++;
    PacketID = numPackets;
    try {
      OTWPcache.put(new Integer(PacketID), this);
    } catch (NullPointerException e) {
      e.printStackTrace();
    }
    
    TLVOptions = new UnsynchVector();
    ClassNames = new Hashtable();
    if (EventMgr == null) {
      EventMgr = (EventManager) 
               ActiveNodeManager.HaveService("EventManager");
      NodeName = ActiveNodeManager.GetNodeName();
    }
    if (EventMgr.traceDepthSupported(TR_MEASURE)) {
      EventMgr.postEvent(TR_STREAM, "ActiveNodeManager.PacketStats.PacketCreation", System.currentTimeMillis(), PacketID);
    }

  }

  /**
   * returns the SmartPacket loader associated with the given type
   * @param type the typeid of the SmartPacket
   * @return SmartPacket loader
   */
  protected static synchronized SmartPacketLoader getLoader(byte[] type) {
    String stype = new String(type);
    SmartPacketLoader loader = (SmartPacketLoader) LoaderCache.get(stype);
    if (loader == null) {
      // System.out.println("OnTheWirepacket: Loader for " + stype + " not in cache");
      loader = new SmartPacketLoader();
      LoaderCache.put(new String(type), loader);
    }
    return loader;
  }

  /**
   * adds the TypeID TLV to the packet
   */
  private void addTypeDef() {

    // create and add a new TLV for this 
    byte flags = (byte) ((1 << 7) & 0x80);  // set Bit 0 to 1 to indicate option is env-specific
    flags |= (byte) 0x0;  // set Bit 1 to 0 to indicate continue processing the header if option unknown
    // append class length info to payload stream
    TLVOption tlv = new TLVOption(flags, (short) TLV_TYPEDEF, TypeID);

    // update the headerlength and packetlength fields
    int tlvLength = tlv.getLength() << 2; // multiply by 4 to get # of bytes
    ANEPheaderLength += tlvLength;
    packetLength += tlvLength;

    // add tlv 
    TLVOptions.addElement(tlv);
  }


  /**
   * adds a new class to the packet
   * @param classname a string naming the class
   * @param ClassData the bytecodes comprising the class definition
   * @exception IOException thrown if the class data cannot be written out as bytes
   */
  protected void addClassDef (String classname, byte[] ClassData) 
                         throws java.io.IOException {

    if (ClassNames.containsKey(classname)) {
      // class def tlv already exists
      return;
    }

    // create and add a new TLV for this class
    byte flags = (byte) ((1 << 7) & 0x80);  // set Bit 0 to 1 to indicate option is env-specific
    flags |= (byte) 0x0;  // set Bit 1 to 0 to indicate continue processing the header if option unknown

    
    byte[] buf = new byte[ClassData.length];
    Deflater def = new Deflater(Deflater.BEST_COMPRESSION);
    def.setInput(ClassData, 0, ClassData.length);
    def.finish();
    int buflen = def.deflate(buf);
    
    // System.err.println("OnTheWirePacket: Class " + classname + " of len " + ClassData.length + " compressed to len " + buflen);
    // append class length info to payload stream
    RefByteArrayOutputStream bos = new RefByteArrayOutputStream(classname.length() + 2 + 4 + buflen);
    DataOutputStream dos = new DataOutputStream(bos);

    dos.writeUTF(classname);
    dos.writeInt(buflen);
    dos.write(buf, 0, buflen);
    if (EventMgr.traceDepthSupported(TR_MEASURE)) {
      String sbuf = "CLASSNAME=" + classname;
      EventMgr.postEvent(TR_STREAM, "ActiveNodeManager.OnTheWirePacket.ClassSent", System.currentTimeMillis(), (long) PacketID, sbuf);      
    }
    TLVOption tlv = new TLVOption(flags, (short) TLV_CLASSDEF, bos.toByteArray());
    bos.close();
    dos.close();

    // update the headerlength and packetlength fields
    int tlvLength = tlv.getLength() << 2; // multiply by 4 to get # of bytes
    ANEPheaderLength += tlvLength;
    packetLength += tlvLength;

    // insert tlv 
    TLVOptions.addElement(tlv);
    ClassNames.put(classname, tlv);
  }


  /** 
   * adds a source address to the packet given the node's virtual name
   * @param nodeName the node's virtual name
   * @exception UnknownHostException if the node name is not a valid active node
   * @see ActiveNetsConstants
   */
  protected void addSourceDef(String nodeName) throws UnknownHostException {
    addAddressDef((short) TLV_SRCIDENT, nodeName);
  }

  /** 
   * adds a destination address to the packet given the node's virtual name
   * @param nodeName the node's virtual name
   *
   * @exception UnknownHostException if the node name is not a valid active node 
   * @see ActiveNetsConstants
   */
  protected void addDestinationDef(String nodeName) throws UnknownHostException {
    addAddressDef((short) TLV_DESTIDENT, nodeName);
  }


  /**
   * adds an address to the packet given the node's virtual name
   * @param option whether source or destination address, constant defined in ActiveNetsConstants
   * @param nodeName the node's virtual name
   * @exception UnknownHostException thrown when the Active Node is not associated with a known host
   *
   * @see ActiveNetsConstants
   */
  private void addAddressDef (short option, String nodeName) throws UnknownHostException {
    // create and add a new TLV for this class
    byte flags = 0x0;
    flags |= (byte) ((1 << 7) & 0x80);  // set Bit 0 to 1 to indicate option is not env-specific
    flags |= (byte) 0x0;  // set Bit 1 to 0 to indicate continue processing the header if option unknown
    
    // convert the virtual name to an IP address
    String physHost = PortInterface.getLinkAddress(nodeName);
    
    byte[] IPv4Address = InetAddress.getByName(physHost).getAddress();
    
    byte[] schemeId = intToBytes(1);
    byte[] payload = new byte[4 + IPv4Address.length];
    // write the scheme identifier first to the payload
    System.arraycopy(schemeId, 0, payload, 0, 4);
    schemeId = null;
    // then copy the address itself
    System.arraycopy(IPv4Address, 0, payload, 4, IPv4Address.length);
    IPv4Address = null;

    TLVOption tlv = new TLVOption(flags, option, payload);

    // update the headerlength and packetlength fields
    int tlvLength = tlv.getLength() << 2; // multiply by 4 to get # of bytes.
    ANEPheaderLength += tlvLength;
    packetLength += tlvLength;
    // add tlv 
    TLVOptions.addElement(tlv);
  }

  /**
   * Set the object for the packet
   * @param obj the executable object which is to be sent in serialized form
   * @exception IOException thrown when the object could not be serialized
   */
  protected void setObject (KUSmartPacketV2 obj) throws IOException {
    Thread thr = Thread.currentThread();
    int prio = thr.getPriority();
    thr.setPriority(prio - 1);
    // don't compute Initial hash here because this host is setting a new object
    // the Initial hash is the received hash (or 0 for new packets) and that will force 
    // WritePacket to actually create byte array for this packet
    SPObject = obj;
    SetObjectID(PacketID);

    //  convert the obj to a stream of bytes
    RefByteArrayOutputStream bos = new RefByteArrayOutputStream();
    SPObjectOutputStream oos = new SPObjectOutputStream (bos, this, SPLoader, SPObject.NextHop);
    oos.writeObject(SPObject); 
    SPLoader.setNodeReached(SPObject.NextHop);

    oos.flush();
    byte[] ObjectData = bos.toByteArray();
    int ObjectDataLength = bos.size();

    /** 
      * hack to gather serialized traces
      *
    try {
      FileOutputStream fos = new FileOutputStream("/users/kulkarn/traces/" + obj.getClass().getName() + ".ser");
      fos.write(ObjectData, 0, ObjectDataLength);
      fos.flush();
      fos.close();
    } catch (Exception fe) {
      fe.printStackTrace();
    }
   ** end hack **/

    oos.close();
    bos.close();
   
    createObjectTLV(ObjectData, ObjectDataLength);
    thr.setPriority(prio);
  }


  /**
   * creates the TLV for the Smartpacket object
   * @param ObjectData byte array representing serialized SmartPacket object
   * @param ObjectDataLength length of the serialized object in bytes
   */
  private void createObjectTLV(byte[] ObjectData, int ObjectDataLength) {  
    /**
    OutgoingHash = ComputeHash(ObjectData);
    **/

    // create a new TLV for the object data
    byte flags = (byte) ((1 << 7) & 0x80);  // set Bit 0 to 1 to indicate option is env-specific
    flags |= 0x0;  // set Bit 1 to 0 to indicate continue processing the header if option unknown

    // append class length info to payload stream
    //    System.out.println("OnTheWirePacket: wrote Object size = " + ObjectData.length);
    byte[] length = intToBytes(ObjectDataLength);
    byte[] payload = new byte[4 + ObjectDataLength];
    System.arraycopy(length, 0, payload, 0, 4);
    System.arraycopy(ObjectData, 0, payload, 4, ObjectDataLength);
    TLVOption tlv = new TLVOption(flags, (short) TLV_OBJECTDEF, payload);
    payload = null;
    ObjectData = null;
    // replace old Object TLV with new one
    if (ObjectTLV != null) {
      // reduce the headerlength and packetlength fields by the size 
      // of the deleted TLV
      int tlvLength = ObjectTLV.getLength() << 2;
      ANEPheaderLength -= tlvLength;
      packetLength -= tlvLength;
      TLVOptions.removeElement(ObjectTLV);
    }

    TLVOptions.addElement(tlv);
    ObjectTLV = tlv;

    // update the headerlength and packetlength fields
    int tlvLength = tlv.getLength() << 2;
    ANEPheaderLength += tlvLength;
    packetLength += tlvLength;
  }

  /**
   * breaks the raw bytes in the ANEP buffer into the correct fields
   * @param packetStream byte array representing on-the-wire SmartPacket
   * @exception StreamCorruptedException if the SmartPacket cannot be reconstructed
   */
  private void assemblePacket(byte[] packetStream) 
    throws StreamCorruptedException {
      try {
        // create a ByteArrayInputStream to read in bytes from buffer

        DataInputStream bufstream = new DataInputStream 
                         (new UnsynchByteArrayInputStream(packetStream));
    
        // version and flags fields are 8 bits long
        version = (short) bufstream.readByte();

        flags = bufstream.readByte();

        // type id field is 16 bits long
        EnvID = new byte[2];
        bufstream.readFully(EnvID);

        // header length field is 16 bits long and specifies the length in 32 bit words
        ANEPheaderLength = (int) bufstream.readShort();
        ANEPheaderLength = ANEPheaderLength << 2; // we save hdr len in bytes

        // total packet length is in octets
        packetLength = (int) bufstream.readUnsignedShort();
        
        int remainingHdrLen = ANEPheaderLength - 8; // 8 bytes indicates header with no TLV options
 
        while (remainingHdrLen > 0) {  // packet contains more TLV options

          long tlvOptTime = System.currentTimeMillis();

          byte[] optField = new byte[2];
          bufstream.readFully(optField);
          // extract the flag and option type bits
          byte flags = (byte) ((optField[0] & 0xC0) >> 6); // get the top two flag bits
          short hiBits = (short) (optField[0] & 0x3F);
          short loBits = (short) (optField[1] & 0xFF);
    
          short optionType = (short) ((hiBits << 8) + (loBits << 0));
          // System.err.println("OnTheWirePacket: TLV Option read = " + optionType);
          // read option length
          int optionLength = (int) bufstream.readShort();
          optionLength *= 4; // since length of option is in 32 bit words
   
          remainingHdrLen -= optionLength;
          // System.err.println("OnTheWirePacket: TLV option length = " + optionLength + " RemHdr = " + remainingHdrLen); 
          if (remainingHdrLen < 0 || optionLength <= 4) throw new StreamCorruptedException();
  
          byte[] optionData = new byte[optionLength - 4];
          bufstream.readFully(optionData, 0, optionData.length); 

          TLVOption tlvOpt = new TLVOption(flags, optionType, optionData);
          TLVOptions.addElement(tlvOpt);
    
          switch (optionType) {
            case TLV_CLASSDEF:
              String classname = ReadClass(optionData);
	      // System.err.println("OnTheWirePacket: Packet " + PacketID + " Class " + classname + " read");
              ClassNames.put(classname, tlvOpt);
              break;
      
            case TLV_OBJECTDEF:
              ReadObject(optionData);
              ObjectTLV = tlvOpt;
              break;
      
            case TLV_TYPEDEF:
              // KU's Typeid is 160 bit hash of the principal class defining the object
              if ((optionLength - 4) != TYPEIDLENGTH) 
                throw new StreamCorruptedException("TypeID is not of correct length");
    
              TypeID = optionData;
              SPLoader = getLoader(TypeID);
              break;
          }

          if (EventMgr.traceDepthSupported(TR_MEASURE)) {
            String sbuf = "VALUE=" + String.valueOf(System.currentTimeMillis() - tlvOptTime);
            EventMgr.postEvent (TR_MEASURE, "ActiveNodeManager.OptionBuild.Interval", 
	  		         System.currentTimeMillis(), PacketID, sbuf);
	  }                    

        }
      } catch (EOFException e) {
        e.printStackTrace();
        throw new StreamCorruptedException();
      } catch (IOException e) {
        e.printStackTrace();
        throw new StreamCorruptedException();
      }
  }       


 /**
  * Read a SmartPacket from the inputstream
  * Embedded objects have their class definitions before the Smartpacket class,
  * so read in all classes before attempting to read object. However, there needs to 
  * be a marker to indicate that the boundary between the class definitions and the
  * serialized object.
  * @param classBytes the bytes comprising the class (including name and length)
  *
  * @return the name of the Class just loaded or null if it cannot be loaded
  * @exception StreamCorruptedException if the bytes do not contain a valid class
  */
  private String ReadClass (byte[] classBytes) 
    throws StreamCorruptedException {
      String ClassName = null;
      try {
        long classReadStart = System.currentTimeMillis();

        DataInputStream spstream = 
             new DataInputStream(new UnsynchByteArrayInputStream(classBytes));
      
        // Read Class Name
        ClassName = spstream.readUTF ();

        // Read the Class
        int ClassLength = spstream.readInt();
      
        String stype = new String(TypeID);

        if (ClassLength > 0) {
	  byte[] ClassBytes = ReadClassData (spstream, ClassLength);

	  SPLoader.makeClass (ClassName, ClassBytes);
          if (EventMgr.traceDepthSupported(TR_STREAM)) {
            String sbuf = "CLASSNAME=" + ClassName;
            EventMgr.postEvent(TR_STREAM, "ActiveNodeManager.OnTheWirePacket.ClassRead", System.currentTimeMillis(), PacketID, sbuf);

            sbuf = "VALUE=" + String.valueOf(System.currentTimeMillis() - classReadStart);
            EventMgr.postEvent (TR_MEASURE, "ActiveNode.ClassBuild.Interval", 
			     System.currentTimeMillis(), (long)PacketID, sbuf);
          }
			  
        }

     } catch (Exception e) {
         e.printStackTrace();
         throw new StreamCorruptedException();
     }
     return ClassName;
  }

  /**
   * reads in the bytes comprising the object and recreates the SmartPacket object
   * @param objectBytes  the bytes making up the object
   *
   * @return true if the object could be deserialized, false otherwise 
   * @exception StreamCorruptedException if the bytes do  not contain a valid object
   */
  private boolean ReadObject(byte[] objectBytes) 
  throws StreamCorruptedException {
    try{
      long deserializeInterval =  System.currentTimeMillis();

      DataInputStream spstream = 
          new DataInputStream(new UnsynchByteArrayInputStream(objectBytes));

      // Read the Object
      int ObjectLength = spstream.readInt();
      //System.err.println("OnTheWirePacket: read in object size = " + ObjectLength);
      if (ObjectLength <= 0) {
        throw new StreamCorruptedException();
      }

      // byte[] objectdata = ReadObjectData (spstream, ObjectLength);

      if (EventMgr.traceDepthSupported(TR_MEASURE)) {
        String sbuf = "VALUE=" + String.valueOf(System.currentTimeMillis() - deserializeInterval);
        EventMgr.postEvent (TR_MEASURE, "ActiveNode.ObjectRead.Interval", 
 			  System.currentTimeMillis(), (long)PacketID, sbuf);
      }
      SPObject = MakeObject (objectBytes, ObjectLength);
      
      if (EventMgr.traceDepthSupported(TR_MEASURE)) {
        String sbuf = "VALUE=" + String.valueOf(System.currentTimeMillis() - deserializeInterval);
        EventMgr.postEvent (TR_MEASURE, "ActiveNode.ObjectMake.Interval", 
 			  System.currentTimeMillis(), (long)PacketID, sbuf);
      }
			  
      /**
      InitialHash = ComputeHash(objectdata);
      **/

      
      if (EventMgr.traceDepthSupported(TR_MEASURE)) {
        String sbuf = "VALUE=" + String.valueOf(System.currentTimeMillis() - deserializeInterval);
        EventMgr.postEvent (TR_MEASURE, "ActiveNode.ObjectBuild.Interval", 
 			  System.currentTimeMillis(), (long)PacketID, sbuf);
      }

      return true;
    } catch (EOFException e) {
       e.printStackTrace();
       throw new StreamCorruptedException();
    } catch (IOException e) {
       e.printStackTrace();
       throw new StreamCorruptedException();
    }
  }
  
  /**
   * Read the length field of a class/object from the input stream.  The class/object is 
   * preceded by its length stored in 4 bytes in big-endian order.
   * @param is the inputstream to read 
   * @return length of the next object in stream
   */
  private int ReadLength (InputStream is) {
    // buffer for length
    byte[] lengthdata = new byte[4];

    // read the length of the class
    try {
      is.read(lengthdata, 0, 4);
    } catch (IOException e) {
      e.printStackTrace();
    }
    // calculate length as an integer
    int ch1 = (int) (lengthdata[0] & 0xFF);
    int ch2 = (int) (lengthdata[1] & 0xFF);
    int ch3 = (int) (lengthdata[2] & 0xFF);
    int ch4 = (int) (lengthdata[3] & 0xFF);
    int length = (int) ((ch1 << 24) + (ch2 << 16) + (ch3 << 8) + 
			(ch4 << 0));
    return length;
  }

  /**
   * write out a short as a byte array
   * @param length the integer (short) to be written out 
   * @return a two byte array repersenting the integer
   */
  private byte[] shortToBytes(short intValue) {
      // write the ANEP header length and packet length fields
      byte length[]  = new byte[2];

      length[0] = (byte) ((intValue >> 8) & 0xFF);
      length[1] = (byte) ((intValue >> 0) & 0xFF);
     
      return length;
  }

  /**
   * write out an int as a byte array
   * @param length the integer (short) to be written out 
   * @return a four byte array representing the integer
   */
  private byte[] intToBytes(int intValue) {
      // write the ANEP header length and packet length fields
      byte length[]  = new byte[4];

      length[0] = (byte) ((intValue >> 24) & 0xFF);
      length[1] = (byte) ((intValue >> 16) & 0xFF);
      length[2] = (byte) ((intValue >> 8) & 0xFF);
      length[3] = (byte) ((intValue >> 0) & 0xFF);
     
      return length;
  }


  // Read a class from the input stream
  private byte[] ReadClassData (DataInputStream is, int length) 
    throws EOFException, IOException {
    // Read the class' description
    byte[] ClassBytes = null;
    try {
      byte[] ClassB = new byte[length];
      is.readFully(ClassB, 0, length);
      Inflater inf = new Inflater();
      inf.setInput(ClassB, 0, length);
      byte [] buf = new byte[length<<2];
      int len = inf.inflate(buf, 0, buf.length);

      // System.err.println("OnTheWirePacket: Class of len " + ClassB.length + " uncompressed to len " + len + " bytes rem = " + inf.getRemaining());
      ClassBytes = new byte[len];
      System.arraycopy(buf, 0, ClassBytes, 0, len);
    } catch (DataFormatException dfe) {
      throw new IOException();
    }
    return ClassBytes;
  }

  /**
   * Read an object from the input stream
   * @param is an inputstream to read the object from
   * @param ObjectLength the number of bytes to read
   * @return the bytes comprising the object
   */
  private byte[] ReadObjectData (DataInputStream is, int ObjectLength) 
  throws EOFException, IOException {  
    // Read the class' description
    byte[] ObjectData = null;
    try {
      ObjectData = new byte[ObjectLength];
      is.readFully (ObjectData, 0, ObjectLength);
    } catch (OutOfMemoryError e) {
      System.err.println("ReadObject: Attempting to allocate " + ObjectLength + " bytes");
      System.exit(-1);
    }
    return ObjectData;
  }

  /**
   * Turn the data into an object
   * @param ObjectData the bytes making up the object
   * @param Objectlength the length of the array in bytes
   * @return a Smartpacket object
   * @exception StreamCorruptedException if the SmartPacket object cannot be reconstructed
   */
  private KUSmartPacketV2 MakeObject (byte[] ObjectData, int Objectlength) 
  throws StreamCorruptedException {
    try {
      UnsynchByteArrayInputStream is = new UnsynchByteArrayInputStream (ObjectData, 4, Objectlength);
      SPObjectInputStream ois = new SPObjectInputStream (is, SPLoader);
      KUSmartPacketV2 obj = (KUSmartPacketV2) ois.readObject();
      return obj;
    } catch (Exception e) {
      e.printStackTrace();
      throw new StreamCorruptedException();
    }
  }

  /**
   * Set the PacketID inside of the Object
   * @param the id to be given to the packet
   */
  private void SetObjectID (int ID) {
    try {
      SPObject.PacketID = ID;
    } catch (NullPointerException e) {
      e.printStackTrace();
    }
  }

  /**
    * computes the hash code for the executable object. 
    * NOTE: you cannot use the Class Object's hashCode() method
    * as explained earlier
    * @param the bytes of the object whose hash is to be computed
    * @return an integer representing a unique hash code
    */
  private int ComputeHash(byte[] objectBytes) {
    // use folding technique to compute hash
   
    long hashTime = System.currentTimeMillis();

    int blk0, blk1, blk2, blk3;
    int hash = 0;
    int nBlocks = (objectBytes.length >> 2) << 2;
    for (int ix = 0; ix < nBlocks; ) {
      // take 4 bytes at  a time
      blk0 = (int) (objectBytes[ix++] & 0xFF);
      blk1 = (int) (objectBytes[ix++] & 0xFF);
      blk2 = (int) (objectBytes[ix++] & 0xFF);
      blk3 = (int) (objectBytes[ix++] & 0xFF);

      int msgBlk = ((blk0 << 24) + (blk1 << 16) + (blk2 << 8) + (blk3 << 0));
 
      hash ^= msgBlk;  // XOR the message block with the previous hash
    }
    // now take care of the excess blocks
    int msgBlk = 0;
    int remBytes = objectBytes.length % 4;  // bytes of message left over
    for (int ix = 0; ix < 4; ix++) {
      int msgByte = 0;
      if (ix < remBytes)
         msgByte = (int) (objectBytes[ix + nBlocks] & 0xFF);
      msgBlk += (int)(msgByte << 8);
    }
    hash ^= msgBlk;

    if (EventMgr.traceDepthSupported(TR_MEASURE)) {
      String sbuf = "VALUE=" + String.valueOf(System.currentTimeMillis() - hashTime);
      EventMgr.postEvent (TR_MEASURE, "ActiveNodeManager.ComputeHash.Interval", 
                        System.currentTimeMillis(), PacketID, sbuf);
    } 

    return hash;
  } 


 /**
  * returns the OTWP object for Smartpacket corresponding to packet id
  *
  * @param packetid the identifier of the Smartpacket
  *
  * @return OnTheWirePacket object or null
  */
  protected static OnTheWirePacket getOTWP(int packetid) {
    if (packetid == 0) 
      return null;
    else
      return (OnTheWirePacket) OTWPcache.get(new Integer(packetid));
  }


 /**
  * removes the OTWP object corresponding to the Smartpacket id from the cache
  *
  * @param packetid the identifier of the Smartpacket
  *
  */
  protected static void destroyOTWP(int packetid) {
    // System.err.println("OnTheWirePacket: Destroying OTWP for packet " + packetid);
    if (packetid != 0) {
      OnTheWirePacket otwp = (OnTheWirePacket) OTWPcache.remove(new Integer(packetid));
      otwp = null;
      
    } else {
      // System.err.println("OnTheWirePacket: Cannot find OTWP for " + packetid);
    }
  }		  


  /**
   * writes the Smartpacket object out as a byte array. 
   * The packet is encapsulated in a ANEP header and tramsmitted
   * as a UDP packet
   * @return a byte array comprising the packet or null if there is an error writing out the packet
   */
  protected byte[] WritePacket () {
    
    Thread thr = Thread.currentThread();
    int prio = thr.getPriority();
    thr.setPriority(prio - 1);

    long packetizeStart = System.currentTimeMillis();

    /***
    // check to see if the SPObject has changed at this node
    if (OutgoingHash == InitialHash) {
      // no need to serialize, just send saved wirestream bytes
      //System.out.println("OnTheWirePacket: Same hash, so sending cached packet");
      return packetStream;
    }
    ***/

    // else recreate wire stream from scratch  
    RefByteArrayOutputStream SPWireBytes = null;

    //create a growable byte array and attach a ObjectOutputStream to it
    SPWireBytes = new RefByteArrayOutputStream(packetLength);

    // Write the version, flags fields
    SPWireBytes.write((byte) ((version >> 0) & 0xFF));
    SPWireBytes.write(flags);

    // write the Env id field in big-endian order 
    SPWireBytes.write(EnvID, 0, 2);

    // write the ANEP header length
    byte[] length = shortToBytes((short) (ANEPheaderLength >> 2));
    SPWireBytes.write (length, 0, 2);
 
    // write the packet length
    length = shortToBytes((short) packetLength);
    // System.out.println("OnTheWirepacket: writing ANEP packet of size = " + packetLength + " header = " + ANEPheaderLength);
    SPWireBytes.write(length, 0, 2);
 
    // write out each tlv option
    for (Enumeration e = TLVOptions.elements(); e.hasMoreElements(); ) {
      // fetch the next TLV's data
      TLVOption tlv = (TLVOption) e.nextElement();
      byte[] tlvData = tlv.toByteArray();
      SPWireBytes.write(tlvData, 0, tlvData.length);
    }
    
    if (TLVOptions.size() > 0) {
      TLVOptions.removeAllElements();
    }
    
    
    // for this env, the payload is null
    
    if (EventMgr.traceDepthSupported(TR_MEASURE)) {
      String sbuf = "VALUE=" + String.valueOf(System.currentTimeMillis() - packetizeStart); 
      EventMgr.postEvent (TR_MEASURE, "ActiveNodeManager.Packetization.Interval", 
			System.currentTimeMillis(), PacketID, sbuf);
    }
                        
    byte[] packetStream = null;
    try {
      packetStream = SPWireBytes.toByteArray();
      SPWireBytes.close();
    } catch (IOException e) {
      packetStream = null;
    }
    SPWireBytes = null;
   
    if (EventMgr.traceDepthSupported(TR_MEASURE)) {
      String sbuf = "VALUE=" + String.valueOf(packetStream.length);
      EventMgr.postEvent (TR_MEASURE, "ActiveNodeManager.Packetization.PacketSize", 
			System.currentTimeMillis(), (long)PacketID, sbuf);
    }

    thr.setPriority(prio);
    return packetStream;
  }

}

