/**

Copyright ( 2001 The University of Kansas
COPYRIGHT AND LIMITATION ON USE
The documents and software (collectively the "Information") developed by the University of Kansas, Information and Telecommunication Technology Center (Hereinafter referred to as KU) is the property of KU or its licensors and is protected by copyright and other intellectual property laws. The information may be viewed, copied, run, modified or printed for educational and research purposes only and shall not be published, sold, broadcast, or redistributed, either directly or indirectly, in whole or in part, by any means or methods for commercial purposes. The Information may not be used for any purpose not expressly approved herein. Uses not approved include, but are not limited to, commercial distribution, in whole or in part.  Any rights not expressly granted herein are reserved by KU. 

Nothing contained herein shall be construed as conferring by implication, estoppel or otherwise any license or right under any patent or trademark of KU or any third party. Except as expressly provided above, nothing contained herein shall be construed as conferring any license or right under any KU copyright. 

DISCLAIMER OF WARRANTY
Use of the  information is at your sole risk. The information, is provided "AS IS" and without warranty of any kind and KU  AND ITS AFFILIATES, AGENTS AND LICENSORS , EXPRESSLY DISCLAIM ALL WARRANTIES, EXPRESS or IMPLIED, INCLUDING, BUT NOT LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND OF FITNESS FOR A PARTICULAR PURPOSE AND ANY EXPRESS OR IMPLIED WARRANTY AGAINST INFRINGEMENT. KU DOES NOT WARRANT THAT THE FUNCTIONS CONTAINED IN THE SOFTWARE WILL MEET YOUR REQUIREMENTS, OR THAT THE OPERATION OF THE SOFTWARE WILL BE UNINTERRUPTED OR ERROR-FREE, OR THAT DEFECTS IN THE SOFTWARE WILL BE CORRECTED. FURTHERMORE, KU DOES NOT WARRANT OR MAKE ANY REPRESENTATIONS REGARDING THE USE OR THE RESULTS OF THE USE OF THE SOFTWARE OR ITS DOCUMENTATION IN TERMS OF THEIR CORRECTNESS, ACCURACY, RELIABILITY, OR OTHERWISE. NO ORAL OR WRITTEN INFORMATION OR ADVICE GIVEN BY KU OR A KU AUTHORIZED REPRESENTATIVE SHALL CREATE A WARRANTY OR IN ANY WAY INCREASE THE SCOPE OF THIS WARRANTY. SHOULD THE SOFTWARE PROVE DEFECTIVE, YOU (AND NOT KU), ASSUMES THE ENTIRE COST OF ALL NECESSARY SERVICING, REPAIR OR CORRECTION. 

LIMITATION OF LIABILITY

NEITHER KU NOR ANY OF ITS AFFILIATES, AGENTS OR REPRESENTATIVES WILL BE LIABLE FOR ACTUAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES (INCLUDING LOSS OF PROFITS, LOSS OF BUSINESS OR SAVINGS), EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGES. IN NO EVENT WILL KU BE LIABLE FOR ANY ACTUAL, SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER, WHETHER IN AN ACTION OF CONTRACT, TORT OR ACTION OF ANY KIND, ARISING OUT OF OR IN CONNECTION WITH THE USE OF THE INFORMATION.  KU, ITS AFFILIATES, AGENTS AND REPRESENTATIVES WILL NOT BE LIABLE FOR NEGLIGENCE, RECKLESSNESS, INTENTIONAL MISCONDUCT OR ANY OTHER ACTS OR OMISSIONS OF THIRD PARTIES. 
GOVERNING LAW
This Agreement, all intellectual property issues, and your rights and obligations shall be governed by the laws of the United States of America and the State of Kansas, excluding their conflicts of laws rules. 


**/

package magician.Node;

import java.io.*;
import java.net.*;
import java.util.*;
import magician.PerfTools.*;

/**
 * The Routing manager is in charge of managing the routing protocols at the
 * active node. The routing manager provides an interface (<tt>RoutingProtocol</tt>
 * for incoming SmartPackets to deploy their own routing strategies. The default 
 * routing protocol, for Smartpackets that do not carry their own, is the 
 * <tt>RIPRouting</tt> protocol.
 * 
 * @see RoutingProtocol
 * @see RIPRouting
 */
public class RoutingMgr {

  /**
   * the routing protocol that is installed as the default
   */
  static RoutingProtocol defaultRouting;

  /**
   * Table of Routing protocols managed
   */
  static Hashtable Schemes;

  /**
   *  Event manager to log events
   */
  static EventManager EventMgr;

  /**
   * constructor sets up the default routing protocol
   */
   public RoutingMgr () {

      Schemes = new Hashtable();

      // crank up the default routing protocol     
      defaultRouting = new RIPRouting();

      EventMgr = (EventManager) ActiveNodeManager.HaveService("EventManager");
   }
   
  /**   
   * called by the active node manager to inform of a new neighbor. This method calls
   * the <tt>PortUp</tt> method of all installed routing protocols to inform them of
   * of the new neighbor
   * @param hadr name of the active node
   */
   public static void PortUp(String hadr){
     for (Enumeration e = Schemes.keys(); e.hasMoreElements(); ) {
        RoutingProtocol rp = (RoutingProtocol) Schemes.get(e.nextElement());
        rp.PortUp(hadr);
     }
     return;

   }
   
  /**
   * called by Active node manager when the active node goes down. This method calls
   * the <tt>PortUp</tt> method of all installed routing protocols to inform them of
   * of the new neighbor
   * @param PortName name of active node
   */
   public static void PortDown(String PortName) {
     for (Enumeration e = Schemes.keys(); e.hasMoreElements(); ) {
        RoutingProtocol rp = (RoutingProtocol) Schemes.get(e.nextElement());
        rp.PortDown(PortName);
     }
   }

   /**
    * returns the default routing protocol
    *
    * @return the default routing protocol
    */
   public static RoutingProtocol defaultRouting() {
     return defaultRouting;
   }


   /**
    * sets the default routing protocol for this node. Can only 
    * be changed by management packets.
    * 
    * @param newProto the new default routing protocol
    * @return the old default
    */
   protected static RoutingProtocol setDefault(RoutingProtocol newProto) {
     RoutingProtocol rp = defaultRouting;
    
     defaultRouting = newProto;
     return rp;
   }


   /**
    * registers a new routing scheme 
    *
    * @param rp the routing protocol being registered
    * @return true if protocol gets registered, false otherwise
    */
   public static boolean Register(RoutingProtocol rp) {
     // obtain the class type
     String key = rp.getClass().getName();

     // store by class type
     Schemes.put(key, rp);
     return true;
   }

}
