/**

Copyright ( 2001 The University of Kansas
COPYRIGHT AND LIMITATION ON USE
The documents and software (collectively the "Information") developed by the University of Kansas, Information and Telecommunication Technology Center (Hereinafter referred to as KU) is the property of KU or its licensors and is protected by copyright and other intellectual property laws. The information may be viewed, copied, run, modified or printed for educational and research purposes only and shall not be published, sold, broadcast, or redistributed, either directly or indirectly, in whole or in part, by any means or methods for commercial purposes. The Information may not be used for any purpose not expressly approved herein. Uses not approved include, but are not limited to, commercial distribution, in whole or in part.  Any rights not expressly granted herein are reserved by KU. 

Nothing contained herein shall be construed as conferring by implication, estoppel or otherwise any license or right under any patent or trademark of KU or any third party. Except as expressly provided above, nothing contained herein shall be construed as conferring any license or right under any KU copyright. 

DISCLAIMER OF WARRANTY
Use of the  information is at your sole risk. The information, is provided "AS IS" and without warranty of any kind and KU  AND ITS AFFILIATES, AGENTS AND LICENSORS , EXPRESSLY DISCLAIM ALL WARRANTIES, EXPRESS or IMPLIED, INCLUDING, BUT NOT LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND OF FITNESS FOR A PARTICULAR PURPOSE AND ANY EXPRESS OR IMPLIED WARRANTY AGAINST INFRINGEMENT. KU DOES NOT WARRANT THAT THE FUNCTIONS CONTAINED IN THE SOFTWARE WILL MEET YOUR REQUIREMENTS, OR THAT THE OPERATION OF THE SOFTWARE WILL BE UNINTERRUPTED OR ERROR-FREE, OR THAT DEFECTS IN THE SOFTWARE WILL BE CORRECTED. FURTHERMORE, KU DOES NOT WARRANT OR MAKE ANY REPRESENTATIONS REGARDING THE USE OR THE RESULTS OF THE USE OF THE SOFTWARE OR ITS DOCUMENTATION IN TERMS OF THEIR CORRECTNESS, ACCURACY, RELIABILITY, OR OTHERWISE. NO ORAL OR WRITTEN INFORMATION OR ADVICE GIVEN BY KU OR A KU AUTHORIZED REPRESENTATIVE SHALL CREATE A WARRANTY OR IN ANY WAY INCREASE THE SCOPE OF THIS WARRANTY. SHOULD THE SOFTWARE PROVE DEFECTIVE, YOU (AND NOT KU), ASSUMES THE ENTIRE COST OF ALL NECESSARY SERVICING, REPAIR OR CORRECTION. 

LIMITATION OF LIABILITY

NEITHER KU NOR ANY OF ITS AFFILIATES, AGENTS OR REPRESENTATIVES WILL BE LIABLE FOR ACTUAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES (INCLUDING LOSS OF PROFITS, LOSS OF BUSINESS OR SAVINGS), EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGES. IN NO EVENT WILL KU BE LIABLE FOR ANY ACTUAL, SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER, WHETHER IN AN ACTION OF CONTRACT, TORT OR ACTION OF ANY KIND, ARISING OUT OF OR IN CONNECTION WITH THE USE OF THE INFORMATION.  KU, ITS AFFILIATES, AGENTS AND REPRESENTATIVES WILL NOT BE LIABLE FOR NEGLIGENCE, RECKLESSNESS, INTENTIONAL MISCONDUCT OR ANY OTHER ACTS OR OMISSIONS OF THIRD PARTIES. 
GOVERNING LAW
This Agreement, all intellectual property issues, and your rights and obligations shall be governed by the laws of the United States of America and the State of Kansas, excluding their conflicts of laws rules. 


**/
package magician.Node;

import java.lang.*;
import java.io.*;
import magician.Utils.*;

/*
 * This class implements the triplets associated with the on-the-wire
 * format of the Smartpacket. The triple is <Type, Length, Value> or
 * TLV, for short.
 * @see OnTheWirePacket
 */
public class TLVOption implements ActiveNetsConstants {
  /**
   * option typeid specific?
   */
  boolean privateOption = false;  // whether the option is TypeID specific

  /**
   * whether to discard packet if option is unrecognized or to ignore it
   * default (false) is ignore
   */
  boolean discardPacket = false;

  /**
   * type of option
   */
  short optionType;  // indicates the type of option
 
  /**
   * length of the entire TLV in 32 bit words
   */
  int optionLength;

  /**
   * value of option
   */
  byte[] value;       // the value field of the option

  /**
   * creates a new TLVOption object of the specified type and value.
   * The flag bits have only two values possible, private and action
   *
   * @param flags  byte representing the flag bits
   * @param type a byte array describing the option type
   * @param val a byte array containing the option
   */
  public TLVOption(byte flags, short type, byte[] val) {
    // bit 8 (lsb) and bit 7 indicate if option is private and 
    // whether to discard packet with unrecognized option
    discardPacket = ((flags & 0x1) == 1) ? true : false;
    privateOption = ((flags & 0x2) == 2) ? true : false;
    optionType = type;
    value = val;
    // optionLength is these many bytes in total including header in 32-bit words, add 3 for padding
    optionLength = ((value.length + 4 + 3)/4);
  }
  
  /**
   * gets the type of the option
   * @return a byte array describing the option type
   */
  public int getOption() {
    return optionType;
  }

  /**
   * gives the total length of the TLV in 32 bit words
   * @return the length of the TLV
   */
  public int getLength() {
    return optionLength;
  }

  /**
   * gives the value of the option
   * @return a byte array containing the value stored
   */
  public byte[] getValue() {
    return value;
  }

  /**
   * tells whether this option is private ie. typeID specific
   * @return true if option is private
   */
  public boolean isPrivate() {
    return privateOption;
  }

  /**
   * tells whether to discard packet if option is unrecognized 
   * or ignore option and continue processing the header
   * @return true if packet is to be discarded and false if TLV is to be ignored
   */
  public boolean discardIfUnrecognized() {
    return discardPacket;
  }

  /**
   * writes the TLV as a byte array suitable for transmission
   * @return a byte array representing the TLV option
   */
  public byte[] toByteArray() {
    RefByteArrayOutputStream out = new RefByteArrayOutputStream(optionLength << 2);
   
    // concatenate the flag and option type values into a 16-bit field
    byte[] optField = new byte [2];
    optField[0] |= (byte) ((privateOption) ? (1 << 7) : 0);
    optField[0] |= (byte) ((discardPacket) ? (1 << 6) : 0);
    optField[0] |= (byte) ((optionType >> 8) & 0x3F);
    out.write(optField[0]);

    optField[1]  = (byte) ((optionType >> 0) & 0xFF);
    out.write(optField[1]);

    // clear the optField bytes to write the option length out
    optField[0] &= 0x0; optField[1] &= 0x0;

    optField[0] = (byte) ((optionLength >> 8) & 0xFF);
    optField[1] = (byte) ((optionLength >> 0) & 0xFF);
    out.write(optField, 0, 2);

    // write out the payload of the option with padding
    //byte[] optionData = new byte[(optionLength - 1) << 2];
    // System.err.println("TLVOption: Option = " + optionType + " value = " + (new String(value)).toString());
    //System.arraycopy(value, 0, optionData, 0, value.length);
    //out.write(optionData, 0, optionData.length);
    out.write(value, 0, value.length);
    int padlength = (optionLength<<2) - 4 - value.length;
    byte[] padding = new byte[padlength];
    out.write(padding, 0, padlength);

    return out.toByteArray();
  }
}
