/**

Copyright ( 2001 The University of Kansas
COPYRIGHT AND LIMITATION ON USE
The documents and software (collectively the "Information") developed by the University of Kansas, Information and Telecommunication Technology Center (Hereinafter referred to as KU) is the property of KU or its licensors and is protected by copyright and other intellectual property laws. The information may be viewed, copied, run, modified or printed for educational and research purposes only and shall not be published, sold, broadcast, or redistributed, either directly or indirectly, in whole or in part, by any means or methods for commercial purposes. The Information may not be used for any purpose not expressly approved herein. Uses not approved include, but are not limited to, commercial distribution, in whole or in part.  Any rights not expressly granted herein are reserved by KU. 

Nothing contained herein shall be construed as conferring by implication, estoppel or otherwise any license or right under any patent or trademark of KU or any third party. Except as expressly provided above, nothing contained herein shall be construed as conferring any license or right under any KU copyright. 

DISCLAIMER OF WARRANTY
Use of the  information is at your sole risk. The information, is provided "AS IS" and without warranty of any kind and KU  AND ITS AFFILIATES, AGENTS AND LICENSORS , EXPRESSLY DISCLAIM ALL WARRANTIES, EXPRESS or IMPLIED, INCLUDING, BUT NOT LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND OF FITNESS FOR A PARTICULAR PURPOSE AND ANY EXPRESS OR IMPLIED WARRANTY AGAINST INFRINGEMENT. KU DOES NOT WARRANT THAT THE FUNCTIONS CONTAINED IN THE SOFTWARE WILL MEET YOUR REQUIREMENTS, OR THAT THE OPERATION OF THE SOFTWARE WILL BE UNINTERRUPTED OR ERROR-FREE, OR THAT DEFECTS IN THE SOFTWARE WILL BE CORRECTED. FURTHERMORE, KU DOES NOT WARRANT OR MAKE ANY REPRESENTATIONS REGARDING THE USE OR THE RESULTS OF THE USE OF THE SOFTWARE OR ITS DOCUMENTATION IN TERMS OF THEIR CORRECTNESS, ACCURACY, RELIABILITY, OR OTHERWISE. NO ORAL OR WRITTEN INFORMATION OR ADVICE GIVEN BY KU OR A KU AUTHORIZED REPRESENTATIVE SHALL CREATE A WARRANTY OR IN ANY WAY INCREASE THE SCOPE OF THIS WARRANTY. SHOULD THE SOFTWARE PROVE DEFECTIVE, YOU (AND NOT KU), ASSUMES THE ENTIRE COST OF ALL NECESSARY SERVICING, REPAIR OR CORRECTION. 

LIMITATION OF LIABILITY

NEITHER KU NOR ANY OF ITS AFFILIATES, AGENTS OR REPRESENTATIVES WILL BE LIABLE FOR ACTUAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES (INCLUDING LOSS OF PROFITS, LOSS OF BUSINESS OR SAVINGS), EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGES. IN NO EVENT WILL KU BE LIABLE FOR ANY ACTUAL, SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER, WHETHER IN AN ACTION OF CONTRACT, TORT OR ACTION OF ANY KIND, ARISING OUT OF OR IN CONNECTION WITH THE USE OF THE INFORMATION.  KU, ITS AFFILIATES, AGENTS AND REPRESENTATIVES WILL NOT BE LIABLE FOR NEGLIGENCE, RECKLESSNESS, INTENTIONAL MISCONDUCT OR ANY OTHER ACTS OR OMISSIONS OF THIRD PARTIES. 
GOVERNING LAW
This Agreement, all intellectual property issues, and your rights and obligations shall be governed by the laws of the United States of America and the State of Kansas, excluding their conflicts of laws rules. 


**/

package magician.PerfTools;

import java.lang.*;
import java.io.*;
import java.util.*;

/**
 * EventManager is a class that manages the event list. Initially, EventManager
 * allocates itself a  circular list of free NetEvents. When an event is posted
 * the next free slot is allocated to the event. 
 * The <tt>writeToOutput()<tt> routine writes
 * all the recorded events to the given output stream.
 *
 * @see NetEvent
 */
 
 
 public class EventManager implements Sleeper, PerfConstants {
    
   /*
    * name of event class
    */
    final static String EVENTCLASS = "magician.PerfTools.NetEvent";

   /*
    * name of list of listeners for all events
    */
    final static String ALL_EVENT_LISTN = "__All_Event_Listeners_";

   /*
    * total number of buffers created 
    */
    protected int size;            

   /*
    * number of buffers in use
    */
    protected int used;           

   /*
    * list of Events which are in use
    */
    protected CircularList EventQ; 

   /*
    * to dump used events every so often
    */
    protected Alarm  alarm;        

   /*
    * list of listeners for events
    */
    protected EventListenerTable eventListeners;

   /*-
    * level of trace detail output to tracestream
    * 0    = System level
    * 1    = kernel stream-level debug
    * 2    = kernel packet-level debug
    * 3    = measurements
    * 4    = reserved
    * 5..  = user level
    */
    protected int tracelevel;
    

  /**
   * creates a pool of free NetEvents
   * @param EventName  name of Event structure being logged
   * @param size   requires size of event pool to be created initially
   *
   * @exception  IllegalArgumentException thrown if  the size of the list is not a positive number
   * @exception  ClassNotFoundException  thrown if the class definition does not exist
   */
    public EventManager(int size) throws IllegalArgumentException, ClassNotFoundException {
      this.size = size;

      // get trace level from environment
      try {
        tracelevel = Integer.parseInt(System.getProperty("TRACELEVEL"));
      } catch (NumberFormatException npe) {
        tracelevel = TR_USER;
      }

      eventListeners = new EventListenerTable();
      used = 0;

      EventQ = new CircularList(Class.forName(EVENTCLASS), size); 
    }

   /** 
    * Classes intending to use periodic flushing of the buffers must call 
    * this method to initiate a timer.
    * @param bufferFlushInterval  the interval in milliseconds
    * @returns  true if buffer flushing is enabled else returns false.
    */
   protected boolean flushBuffers(long bufferFlushInterval) {
     // set alarm to callback every bufferFlushInterval milliseconds
     if (bufferFlushInterval > 0) {
        alarm = new Alarm(this, bufferFlushInterval);
        alarm.setPriority(Thread.MAX_PRIORITY - 2);
        alarm.start();
        Thread.currentThread().yield();
        return true;
     } 
     else return false;
   }


  /**
   * returns the next free event in the circular list. Also does house keeping.
   */
   protected NetEvent getFreeEvent() {
     // remove a free NetEvent from the FreeList 
     NetEvent ev = null;
     try {
       ev = (NetEvent) EventQ.set();
     } catch (NullPointerException e) {
         // buffers are full, call dump routine to free buffers
         writeToOutput();
         ev = (NetEvent) EventQ.set();
     }
     used++;
     return ev;
   }

   /**
    * check if event at particular depth should be logged
    *
    * @param depth trace level for event
    * @returns true if event can be logged
    */
    public boolean traceDepthSupported(int depth) {
      return (tracelevel > depth)? false: true;
    }


   /**
    *  posts an event by calling the NetEvent's update method (if provided) 
    *  to record event information,
    *  else declare the variables of the NetEvent class as public and 
    *  update them directly in this method.
    *  Classes implementing this method must call getFreeEvent() to get 
    *  the NetEvent object to update.
    */
    public synchronized void postEvent(int traceDepth, String eventName, long eventTime, long packetID, String otherParams) {
	// get free NetEvent object and update it
	NetEvent ev = getFreeEvent();
	ev.updateEventFields(traceDepth, eventName, eventTime, packetID, otherParams);

	// notify synchronous listeners interested in this event
	String eventType = eventName + "_" + String.valueOf(true);
	Vector list = eventListeners.getListenersForType(eventType);
	dispatchEvent(ev, list);
    }

   /**
    *  posts an event by calling the NetEvent's update method (if provided) 
    *  to record event information,
    *  else declare the variables of the NetEvent class as public and 
    *  update them directly in this method.
    *  Classes implementing this method must call getFreeEvent() to get 
    *  the NetEvent object to update.
    */
    public synchronized void postEvent(int traceDepth, String eventName, long eventTime, long packetID) {
	// get free NetEvent object and update it
	NetEvent ev = getFreeEvent();
	ev.updateEventFields(traceDepth, eventName, eventTime, packetID);

	// notify synchronous listeners interested in this event
	String eventType = eventName + "_" + String.valueOf(true);
	Vector list = eventListeners.getListenersForType(eventType);
        dispatchEvent(ev, list);
    }

   /**
    * calls internal dump routine to write records to output stream. 
    */
    public void wakeup() {
       writeToOutput();
    }

  /**
   * adds listener for specific events generated by EventManager.
   * Listener must conform to the <tt>NetEventListener</tt> interface.
   * The listener can choose to be notified synchronously or asynchronously.
   * Typically, synchronous notification is to be used for alarms while
   * asynchronous notification is to be used for normal reporting.
   *
   * @param EventName event to listen 
   * @param listener name of event listener
   * @param syncFlag set to true if notification must be immediate
   * @see NetEventListener
   */
   public void addListenerForEvent(String EventName, NetEventListener listener, boolean syncFlag) {
       // get list of listeners for that event
       String eventType = EventName + "_" + String.valueOf(syncFlag);
       eventListeners.add(eventType, listener);
   }

  /**
   * adds listener for all events generated by EventManager.
   * Listener must conform to the <tt>NetEventListener</tt> interface.
   * The listener can choose to be notified synchronously or asynchronously.
   * Typically, synchronous notification is to be used for alarms while
   * asynchronous notification is to be used for normal reporting.
   *
   * @param listener name of event listener
   * @param syncFlag set to true if notification must be immediate
   * @see NetEventListener
   */
   public void addListenerForAllEvents(NetEventListener listener, boolean syncFlag) {
       // get list of listeners for that event
       String eventType = ALL_EVENT_LISTN + "_" + String.valueOf(syncFlag);
       eventListeners.add(eventType, listener);
   }

  /**
   * removes listener from specific events generated by EventManager.
   * Listener must conform to the <tt>NetEventListener</tt> interface
   *
   * @param EventName name of event 
   * @param listener name of event listener
   * @see NetEventListener
   */
   public boolean removeListenerForEvent(String EventName, NetEventListener listener) {
       boolean removed = false;

       // get list of listeners for that event
       String eventType = EventName + "_" + String.valueOf(true);
       removed = eventListeners.remove(eventType, listener);

       if (! removed) {
	   // must be in async list
	   eventType = EventName + "_" + String.valueOf(false);
	   removed = eventListeners.remove(eventType, listener);
       }

       return removed;
   }

  /**
   * removes listener registered for all events generated by EventManager.
   * Listener must conform to the <tt>NetEventListener</tt> interface
   *
   * @param EventName name of event 
   * @param listener name of event listener
   * @see NetEventListener
   */
   public boolean removeListenerForAllEvents(NetEventListener listener) {
       return removeListenerForEvent(ALL_EVENT_LISTN, listener);
   }


   /**
    * writes out all the NetEvent objects to all listeners
    */
    protected synchronized void writeToOutput() {
      // send events to all async listeners
      NetEvent ev = null;
      while ((ev = (NetEvent) EventQ.get()) != null) {
	  // get async listeners for this event
          String eventType = ev.getEventType() + "_" + String.valueOf(false);
          Vector list = (Vector) eventListeners.getListenersForType(eventType);
	  dispatchEvent(ev, list);
          eventType = ALL_EVENT_LISTN + "_" + String.valueOf(false);
          list = (Vector) eventListeners.getListenersForType(eventType);
	  dispatchEvent(ev, list);
          used--;
      }
    } 

   /**
    * fire off signal to listeners that event has occured
    */
     private void dispatchEvent(NetEvent ev, Vector list) {
         if (list == null || list.size() == 0) return;

	 for (Enumeration e = list.elements(); e.hasMoreElements(); ) {
	     NetEventListener el = (NetEventListener) e.nextElement();
	     el.postEvent(ev);
	 }
     }


   /**
    * stops the EventManager and prints out all records to the output stream.
    */
   public void stop() {
     writeToOutput();
     // stop the alarm too if it was set
     if (alarm != null) {
       alarm.cancel();
       alarm.stop();
     }
   }


 }
