/**

Copyright ( 2001 The University of Kansas
COPYRIGHT AND LIMITATION ON USE
The documents and software (collectively the "Information") developed by the University of Kansas, Information and Telecommunication Technology Center (Hereinafter referred to as KU) is the property of KU or its licensors and is protected by copyright and other intellectual property laws. The information may be viewed, copied, run, modified or printed for educational and research purposes only and shall not be published, sold, broadcast, or redistributed, either directly or indirectly, in whole or in part, by any means or methods for commercial purposes. The Information may not be used for any purpose not expressly approved herein. Uses not approved include, but are not limited to, commercial distribution, in whole or in part.  Any rights not expressly granted herein are reserved by KU. 

Nothing contained herein shall be construed as conferring by implication, estoppel or otherwise any license or right under any patent or trademark of KU or any third party. Except as expressly provided above, nothing contained herein shall be construed as conferring any license or right under any KU copyright. 

DISCLAIMER OF WARRANTY
Use of the  information is at your sole risk. The information, is provided "AS IS" and without warranty of any kind and KU  AND ITS AFFILIATES, AGENTS AND LICENSORS , EXPRESSLY DISCLAIM ALL WARRANTIES, EXPRESS or IMPLIED, INCLUDING, BUT NOT LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND OF FITNESS FOR A PARTICULAR PURPOSE AND ANY EXPRESS OR IMPLIED WARRANTY AGAINST INFRINGEMENT. KU DOES NOT WARRANT THAT THE FUNCTIONS CONTAINED IN THE SOFTWARE WILL MEET YOUR REQUIREMENTS, OR THAT THE OPERATION OF THE SOFTWARE WILL BE UNINTERRUPTED OR ERROR-FREE, OR THAT DEFECTS IN THE SOFTWARE WILL BE CORRECTED. FURTHERMORE, KU DOES NOT WARRANT OR MAKE ANY REPRESENTATIONS REGARDING THE USE OR THE RESULTS OF THE USE OF THE SOFTWARE OR ITS DOCUMENTATION IN TERMS OF THEIR CORRECTNESS, ACCURACY, RELIABILITY, OR OTHERWISE. NO ORAL OR WRITTEN INFORMATION OR ADVICE GIVEN BY KU OR A KU AUTHORIZED REPRESENTATIVE SHALL CREATE A WARRANTY OR IN ANY WAY INCREASE THE SCOPE OF THIS WARRANTY. SHOULD THE SOFTWARE PROVE DEFECTIVE, YOU (AND NOT KU), ASSUMES THE ENTIRE COST OF ALL NECESSARY SERVICING, REPAIR OR CORRECTION. 

LIMITATION OF LIABILITY

NEITHER KU NOR ANY OF ITS AFFILIATES, AGENTS OR REPRESENTATIVES WILL BE LIABLE FOR ACTUAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES (INCLUDING LOSS OF PROFITS, LOSS OF BUSINESS OR SAVINGS), EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGES. IN NO EVENT WILL KU BE LIABLE FOR ANY ACTUAL, SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER, WHETHER IN AN ACTION OF CONTRACT, TORT OR ACTION OF ANY KIND, ARISING OUT OF OR IN CONNECTION WITH THE USE OF THE INFORMATION.  KU, ITS AFFILIATES, AGENTS AND REPRESENTATIVES WILL NOT BE LIABLE FOR NEGLIGENCE, RECKLESSNESS, INTENTIONAL MISCONDUCT OR ANY OTHER ACTS OR OMISSIONS OF THIRD PARTIES. 
GOVERNING LAW
This Agreement, all intellectual property issues, and your rights and obligations shall be governed by the laws of the United States of America and the State of Kansas, excluding their conflicts of laws rules. 


**/
/*
 * @(#)DatagramPacketPool.java	1.12 97/01/25
 * 
 */


package magician.Utils;

import java.util.*;
import java.net.*;

public 
class DatagramPacketPool {

    final static int MAX_DATAGRAM_PACKET_LEN = 65536;

    /*
     * a pool of null-initialialized instances of this class stored in a
     * Vector for rapid access. default size of pool is 8
     */
    static Vector packetPool; 

    /**
     * Constructs a <code>DatagramPacket</code> for receiving packets of 
     * length <code>ilength</code>. 
     * <p>
     * The <code>length</code> argument must be less than or equal to 
     * <code>ibuf.length</code>. 
     *
     * @param   psize   the number of initialized datagram packets in the pool.
     * @since   JDK1.1
     */
    private DatagramPacketPool(int psize) throws IllegalArgumentException {
	if (psize <= 0) {
	    throw new IllegalArgumentException("size must be positive");
	}
        packetPool = new Vector(psize);
    }
    

    /**
     * returns an initialized instance of a UnsynchDatagramPacket
     *
     * @return a DatagramPacket
     */
    public synchronized static DatagramPacket getNewDatagramPacket() {
      DatagramPacket packet = null;
      try {
        packet = (DatagramPacket) packetPool.elementAt(0);
        packetPool.removeElementAt(0);
      } catch (NullPointerException e) {
         // packetPool not created
         // we choose size 4 because there are 4 UDP ports for this version
         // and hence 4 port managers who can be requesting at any time.
         new DatagramPacketPool(4);
         int psize = 4;
         while (psize > 0) {
           byte [] data = new byte[MAX_DATAGRAM_PACKET_LEN];
           packetPool.addElement(new DatagramPacket(data, data.length));
           psize--;
         }
         packet = (DatagramPacket) packetPool.elementAt(0);
         packetPool.removeElementAt(0);
      } catch (ArrayIndexOutOfBoundsException e) {
         // no empty packets left in pool
         int size = packetPool.capacity();
         int newsize = size << 1;
         packetPool.ensureCapacity(newsize);

         while (newsize - size > 0) {
           byte [] data = new byte[MAX_DATAGRAM_PACKET_LEN];
           packetPool.addElement(new DatagramPacket(data, data.length));
           size++;
         }
         packet = (DatagramPacket) packetPool.elementAt(0);
         packetPool.removeElementAt(0);
      }
      return packet;
    }

    /**
     * release a datagram packet instance back to the free pool
     *
     * @param packet  the datagram packet to be returned
     * @exception IllegalAccessException if free is called before allocating DatagramPacket 
     */
    public synchronized static void free(DatagramPacket packet) throws IllegalAccessException {
      try {
        reinitialize(packet);
        packetPool.addElement(packet);
      } catch (NullPointerException e) {
          throw new IllegalAccessException("free cannot be called before getNewDatagramPacket method");
      }
    }

    /**
     * returns the size of the Datagram packet pool
     *
     * @return the size of the pool
     */
    public static int size() {
      return packetPool.size();
    }

    /**
     * re-initializes the packet for reuse
     * idea taken from Singhal etal paper from IBM 
     * TJ Watson Research Center
     */
    private static void reinitialize(DatagramPacket dp) {
      byte[] data = new byte[MAX_DATAGRAM_PACKET_LEN];
      dp.setData(data);
      dp.setLength(data.length);
    }
}
